#!/usr/bin/env python

from ..surveyor import Surveyor
from . import Explorer

import logging
l = logging.getLogger("angr.surveyors.Escaper")

class Escaper(Surveyor):
	'''
	Escaper implements loop escaping!

		normal - any found normal paths from the loop
		forced - forced paths from the loop, if a normal wasn't found
	'''

	def __init__(self, project, loop_addresses, start=None, max_concurrency=None, max_active=None, pickle_paths=None, loop_iterations=0, iteration_depth=100, unconstrain_memory=True, unconstrain_registers=True):
		'''
		Creates an Escaper. Most options are for Surveyor (separate docs).

		@param loop_addresses: the addresses of all the basic blocks in the loop, to know the
							   instructions to which the analysis should be restricted
		@param loop_iterations: the number of times to run the loop before escaping
		@param iteration_depth: the maximum depth (in SimRuns) of a path through the loop
		'''
		Surveyor.__init__(self, project, start=start, max_concurrency=max_concurrency, max_active=max_active, pickle_paths=pickle_paths)

		self._loop_addresses = loop_addresses
		self._loop_iterations = loop_iterations
		self._iteration_depth = iteration_depth
		self._current_iteration = 0
		self._done = False

		self._unconstrain_memory = unconstrain_memory
		self._unconstrain_registers = unconstrain_registers

		self.normal = [ ]
		self.forced = [ ]

	def _tick_loop(self, start=None):
		results = Explorer(self._project, start=start, find=self._loop_addresses[0], restrict=self._loop_addresses, min_depth=2, max_depth=self._iteration_depth, max_repeats=1, max_concurrency=self._max_concurrency, num_find=self._num_find).run()

		self.deadended += results.deadended
		return results

	def unconstrain_loop(self, constrained_entry):
		'''
		Unconstrains an exit to the loop header by looping one more time
		and replacing all modified variables with unconstrained versions.
		'''

		constrained_state = constrained_entry.state.copy()

		# first, go through the loop normally, one more time
		constrained_results = self._tick_loop(start=constrained_entry)
		l.debug("%d paths to header found", len(constrained_results.found))

		# then unconstrain differences between the original state and any new
		# head states
		unconstrained_states = []
		for p in constrained_results.found:
			# because the head_entry might actually point partway *through* the
			# loop header, in the cases of a loop starting between
			# the counter-increment and the condition check (because the
			# counter is only incremented at the end of the loop, and the
			# end is placed in the beginning for optimization), so we run the
			# loop through to the *end* of the header
			new_state = p.state.copy()
			if self._unconstrain_registers:
				new_state.registers.unconstrain_differences(constrained_state.registers)
			if self._unconstrain_memory:
				new_state.memory.unconstrain_differences(constrained_state.memory)

			unconstrained_states.append(new_state)
		l.debug("%d unconstrained states", len(unconstrained_states))

		unconstrained_exits = []
		unconstrained_entry = constrained_entry
		for s in unconstrained_states:
			unconstrained_entry.state = s
			unconstrained_results = self._tick_loop(start=unconstrained_entry)

			unconstrained_exits += unconstrained_results.deviating

		return unconstrained_exits

	def tick(self):
		'''
		Makes one run through the loop.
		'''
		if self._current_iteration < self._loop_iterations:
			l.debug("Currently at iteration %d of %d", self._current_iteration, self._loop_iterations)

			results = self._tick_loop(start=self.active_exits(reachable=True))

			l.debug("... found %d exiting paths", len(results.deviating))
			self.normal += results.deviating
			self.active = results.found

			self._current_iteration += 1
		else:
			all_exits = self.active_exits(reachable=True)
			l.debug("Unconstraining %d heads.", len(all_exits))
			for e in all_exits:
				self.forced += self.unconstrain_loop(e)

			self._done = True

	@property
	def done(self):
		return self._done

	def __repr__(self):
		return "<Escaper with paths: %s, %d normal, %d forced>" % (Surveyor.__repr__(self), len(self.normal), len(self.forced))

from . import all_surveyors
all_surveyors['Escaper'] = Escaper



import copy
import logging
from itertools import dropwhile

import simuvex

l = logging.getLogger(name="angr.entry_wrapper")

class CallStack(object):
    def __init__(self, stack=None, retn_targets=None, stack_pointers=None, accessed_registers=None):
        self._stack = [ ] if stack is None else stack
        self._retn_targets = [ ] if retn_targets is None else retn_targets
        self._stack_pointers = [ ] if stack_pointers is None else stack_pointers
        self._accessed_registers = [ ] if accessed_registers is None else accessed_registers

    def __len__(self):
        """
        Get how many functions calls there are in the current stack
        :return:
        """

        return len(self._stack)

    def __repr__(self):
        """
        :return: A proper representation of the CallStack object
        """
        return "<CallStack of %d frames>" % len(self._stack)

    def dbg_print(self):
        """
        :return: Details of this CalLStack
        """

        stack = [ ]
        for i, (st, return_target) in enumerate(reversed(zip(self._stack, self._retn_targets))):
            # Unpack the tuple
            callsite_irsb_addr, func_addr = st

            s = "%d | %s -> %s, returning to %s" % (
                i,
                "None" if callsite_irsb_addr is None else hex(callsite_irsb_addr),
                "None" if func_addr is None else hex(func_addr),
                "None" if return_target is None else hex(return_target)
            )
            stack.append(s)

        return "\n".join(stack)

    def clear(self):
        self._stack = [ ]
        self._retn_targets = [ ]
        self._stack_pointers = [ ]

    @staticmethod
    def stack_suffix_to_string(stack_suffix):
        '''
        Convert a stack suffix to a human-readable string representation.
        :param stack_suffix:
        :return: A string
        '''
        s = "[" + ",".join([("0x%x" % i) if i is not None else "Unspecified" for i in stack_suffix]) + "]"
        return s

    def stack_suffix(self, context_sensitivity_level):
        length = len(self._stack)

        ret = ()
        for i in xrange(context_sensitivity_level):
            index = length - i - 1
            if index < 0:
                ret = (None, ) + ret
            else:
                ret = self._stack[index] + ret
        return ret

    def call(self, callsite_addr, addr, retn_target=None, stack_pointer=None):
        self._stack.append((callsite_addr, addr))
        self._retn_targets.append(retn_target)
        self._stack_pointers.append(stack_pointer)
        self._accessed_registers.append(set())

    @property
    def current_function_address(self):
        if len(self._stack) == 0:
            return 0 # This is the root level
        else:
            return self._stack[-1][-1]

    @property
    def all_function_addresses(self):
        """
        Get all function addresses called in the path, from the earliest one to the most recent one
        :return: a list of function addresses
        """
        return [ s[-1] for s in self._stack ]

    @property
    def current_stack_pointer(self):
        if len(self._stack) == 0:
            return None
        else:
            return self._stack_pointers[-1]

    @property
    def current_function_accessed_registers(self):
        if len(self._accessed_registers) == 0:
            return set()
        else:
            return self._accessed_registers[-1]

    @staticmethod
    def _rfind(lst, item):
        try:
            return dropwhile(lambda x: lst[x] != item,
                             reversed(xrange(len(lst)))).next()
        except Exception:
            raise ValueError("%s not in the list" % item)

    def ret(self, retn_target):
        if retn_target in self._retn_targets:
            # We may want to return to several levels up there, not only a
            # single stack frame
            levels = len(self._retn_targets) - \
                self._rfind(self._retn_targets, retn_target)
        else:
            l.warning("Returning to unexpected address 0x%08x", retn_target)
            # For Debugging
            # raise Exception()
            # There are cases especially in ARM where return is used as a jump
            # So we don't pop anything out
            levels = 0
        while levels > 0:
            if len(self._stack) > 0:
                self._stack.pop()
            if len(self._retn_targets) > 0:
                self._retn_targets.pop()
            if len(self._stack_pointers) > 0:
                self._stack_pointers.pop()
            if len(self._accessed_registers) > 0:
                self._accessed_registers.pop()
            levels -= 1

    def get_ret_target(self):
        if len(self._retn_targets) == 0:
            return None
        return self._retn_targets[-1]

    def copy(self):
        return CallStack(self._stack[::], self._retn_targets[::], self._stack_pointers[::], self._accessed_registers[::])

class BBLStack(object):
    def __init__(self, stack_dict=None):
        if stack_dict is None:
            self._stack_dict = { }
        else:
            self._stack_dict = stack_dict

    @staticmethod
    def _get_key(callstack_suffix, func_addr):
        if len(callstack_suffix) > 0:
            key = callstack_suffix
        else:
            key = func_addr

        return key

    def copy(self):
        return BBLStack(copy.deepcopy(self._stack_dict))

    def call(self, callstack_suffix, func_addr):
        key = self._get_key(callstack_suffix, func_addr)

        # Create a stack with respect to that function
        self._stack_dict[key] = []

    def ret(self, callstack_suffix, func_addr):
        key = self._get_key(callstack_suffix, func_addr)

        if key in self._stack_dict:
            # Return from a function. Remove the corresponding stack
            del self._stack_dict[key]
        else:
            l.warning("Attempting to ret from a non-existing stack frame %s.", hex(key) if isinstance(key, (int, long)) else key)

    def push(self, callstack_suffix, func_addr, bbl):
        key = self._get_key(callstack_suffix, func_addr)

        if key not in self._stack_dict:
            l.warning("Key %s is not in stack dict. It might be caused by " + \
                      "an unexpected exit target.", hex(key) if isinstance(key, (int, long)) else key)
            self.call(callstack_suffix, func_addr)
        self._stack_dict[key].append(bbl)

    def in_stack(self, callstack_suffix, func_addr, bbl):
        key = self._get_key(callstack_suffix, func_addr)

        if key in self._stack_dict:
            return bbl in self._stack_dict[key]
        return False

    def __repr__(self):
        s = [ ]
        for key, stack in self._stack_dict.iteritems():
            s_ = ", ".join([ (hex(k) if k is not None else "None") for k in key ])
            s_ = "[" + s_ + "]:\n  "
            s_ += " -> ".join([ hex(k) for k in stack ])

            s.append(s_)

        return "\n".join(s)

class EntryWrapper(object):
    def __init__(self, path, context_sensitivity_level, call_stack=None, bbl_stack=None, is_narrowing=False):
        self._path = path

        # Other parameters
        self._context_sensitivity_level = context_sensitivity_level
        self.is_narrowing = is_narrowing

        if call_stack is None:
            self._call_stack = CallStack()

            # Added the function address of the current exit to callstack
            se = self._path.state.se
            sp_expr = self._path.state.regs.sp

            # If the sp_expr cannot be concretized, the stack pointer cannot be traced anymore.
            try:
                sp = se.exactly_n_int(sp_expr, 1)[0]
            except (simuvex.SimValueError, simuvex.SimSolverModeError):
                l.warning("Stack pointer cannot be concretized. CallStack cannot track the stack pointer changes.")

                # Set the stack pointer to None
                sp = None

            self._call_stack.call(None, self._path.addr, stack_pointer=sp)

            self._bbl_stack = BBLStack()
            # Initialize the BBL stack
            self._bbl_stack.call(self._call_stack.stack_suffix(self._context_sensitivity_level), path.addr)
        else:
            self._call_stack = call_stack
            self._bbl_stack = bbl_stack
        assert self._call_stack is not None and self._bbl_stack is not None

    @property
    def path(self):
        return self._path

    @property
    def call_stack(self):
        return self._call_stack

    def call_stack_copy(self):
        return self._call_stack.copy()

    def call_stack_suffix(self):
        return self._call_stack.stack_suffix(self._context_sensitivity_level)

    def bbl_stack_push(self, call_stack_suffix, function_addr, bbl_addr):
        self._bbl_stack.push(call_stack_suffix, function_addr, bbl_addr)

    def bbl_in_stack(self, call_stack_suffix, function_addr, bbl_addr):
        return self._bbl_stack.in_stack(call_stack_suffix, function_addr, bbl_addr)

    def bbl_stack(self):
        return self._bbl_stack

    def bbl_stack_copy(self):
        return self._bbl_stack.copy()

    @property
    def current_function_address(self):
        return self._call_stack.current_function_address

    @property
    def current_stack_pointer(self):
        return self._call_stack.current_stack_pointer

    @property
    def current_function_accessed_registers(self):
        return self._call_stack.current_function_accessed_registers



from collections import defaultdict
import copy

from simuvex import SimIRSB, SimProcedure
from ..analysis import Analysis

import logging

l = logging.getLogger("angr.analyses.ddg")

MAX_ANALYZE_TIMES = 3

class DDG(Analysis):
    def __init__(self, cfg, start=None):
        self._project = self._p
        self._cfg = cfg
        self._entry_point = self._project.entry

        self._graph = defaultdict(lambda: defaultdict(set))
        self._symbolic_mem_ops = set()

        self._construct()

    @property
    def graph(self):
        return self._graph

    def pp(self):
        for k,v in self.graph.iteritems():
            for st, tup in v.iteritems():
                print "(0x%x, %d) <- (0x%x, %d)" % (k, st,
                                                            list(tup)[0][0],
                                                            list(tup)[0][1])

    def debug_print(self):
        l.debug(self._graph)

    def _trace_source(self, init_run, init_ref):
        '''
        Trace the sources (producers) of a given ref
        '''
        l.debug("Tracing source for symbolic memory dep %s of %s", init_ref, init_run)
        sources = set()
        # Memorization
        traced = {}

        init_stmt_id = init_ref.stmt_idx
        init_reg_deps = set()
        init_tmp_deps = set()
        real_ref = init_run.statements[init_stmt_id].refs[-1]
        if type(real_ref) == SimMemWrite:
            init_reg_deps |= set(list(real_ref.data_reg_deps))
            init_tmp_deps |= set(list(real_ref.data_tmp_deps))
            init_reg_deps |= set(list(real_ref.addr_reg_deps))
            init_tmp_deps |= set(list(real_ref.addr_tmp_deps))
        elif type(real_ref) == SimMemRead:
            init_reg_deps |= set(list(real_ref.addr_reg_deps))
            init_tmp_deps |= set(list(real_ref.addr_tmp_deps))
        else:
            init_reg_deps |= set(list(real_ref.data_reg_deps))
            init_tmp_deps |= set(list(real_ref.data_tmp_deps))
        stack = [(init_run, init_stmt_id, init_reg_deps, init_tmp_deps)]
        while len(stack) > 0:
            run, stmt_id, reg_deps, tmp_deps = stack.pop()
            # l.debug("Traversing %s", run)
            # Added to traced set
            traced[run.addr] = reg_deps
            reg_dep_to_stmt_id = {}

            if isinstance(run, SimIRSB):
                irsb = run
                if stmt_id == -1:
                    stmt_ids = range(len(irsb.statements))
                else:
                    stmt_ids = range(stmt_id + 1)
                stmt_ids.reverse()

                for stmt_id in stmt_ids:
                    stmt = irsb.statements[stmt_id]
                    if len(stmt.refs) > 0:
                        real_ref = stmt.refs[-1]
                        if type(real_ref) == SimRegWrite and real_ref.offset in reg_deps:
                            reg_dep_to_stmt_id[real_ref.offset] = stmt_id
                            reg_deps.remove(real_ref.offset)
                            reg_deps |= set(list(real_ref.data_reg_deps))
                            tmp_deps |= set(list(real_ref.data_tmp_deps))
                        elif type(real_ref) == SimTmpWrite and real_ref.tmp in tmp_deps:
                            tmp_deps.remove(real_ref.tmp)
                            reg_deps |= set(list(real_ref.data_reg_deps))
                            tmp_deps |= set(list(real_ref.data_tmp_deps))
            elif isinstance(run, SimProcedure):
                refs = run.refs()
                refs.reverse()
                for ref in refs:
                    if type(ref) == SimRegWrite and ref.offset in reg_deps:
                        reg_dep_to_stmt_id[ref.offset] = -1
                        reg_deps.remove(ref.offset)
                        reg_deps |= set(list(ref.data_reg_deps))
                        tmp_deps |= set(list(ref.data_tmp_deps))
                    elif type(ref) == SimTmpWrite and ref.tmp in tmp_deps:
                        tmp_deps.remove(ref.tmp)
                        reg_deps |= set(list(ref.data_reg_deps))
                        tmp_deps |= set(list(ref.data_tmp_deps))

            if len(reg_deps) > 0:
                predecessors = self._cfg.get_predecessors(run)
                for s in predecessors:
                    # tpl = (s.addr, reg_deps)
                    if s.addr not in traced:
                        stack.append((s, -1, reg_deps.copy(), set()))
                    else:
                        old_reg_deps = traced[s.addr]
                        if not reg_deps.issubset(old_reg_deps):
                            new_reg_deps = reg_deps.copy()
                            new_reg_deps |= old_reg_deps
                            stack.append((s, -1, new_reg_deps, set()))
            for reg, stmt_id in reg_dep_to_stmt_id.items():
                sources.add((run.addr, stmt_id))
            if len(reg_deps) == 0 or len(predecessors) == 0:
                # This is the end of the u-d chain
                # FIXME: What if there is a loop, and len(successors) never == 0?
                for reg in reg_deps:
                    if reg not in reg_dep_to_stmt_id:
                        # It's never been written before
                        # We use a fake run.addr and stmt_id so that we don't lose them
                        l.debug("Register %d has never been assigned a value before.", reg)
                        sources.add((-1 * reg, -1))
                    else:
                        # sources.add((run.addr, reg_dep_to_stmt_id[reg]))
                        # They have already been added to sources
                        pass

        return sources

    def _solve_symbolic_mem_operations(self):
        '''
        We try to resolve symbolic memory operations in the following manner:
        For each memory operation, trace the pointer from its root producers.
        And then relate each memory read with all memory writes that has shared
        producers with it.
        It's imprecise and could be over-approximating, but it's better than
        losing dependencies.
        '''
        # irsb, stmt_id => list of tuples (irsb, ref)
        mem_read_producers = defaultdict(list)
        mem_write_producers = defaultdict(list)
        for irsb, ref in self._symbolic_mem_ops:
            sources = self._trace_source(irsb, ref)
            for src_irsb_addr, stmt_id in sources:
                if type(ref) == SimMemRead:
                    mem_read_producers[(src_irsb_addr, stmt_id)].append((irsb.addr, ref))
                elif type(ref) == SimMemWrite:
                    mem_write_producers[(src_irsb_addr, stmt_id)].append((irsb.addr, ref))

        for tpl, lst_writes in mem_write_producers.items():
            if tpl in mem_read_producers:
                lst_reads = mem_read_producers[tpl]
                for read_irsb, read_ref in lst_reads:
                    for write_irsb, write_ref in lst_writes:
                        self._graph[read_irsb][read_ref.stmt_idx].add((write_irsb, write_ref.stmt_idx))

    def _construct(self):
        '''
        We track the following types of memory access:
        - (Intra-functional) Stack read/write.
            Trace changes of stack pointers inside a function, and the dereferences
            of stack pointers.
        - (Inter-functional) Stack read/write.
            TODO
        - (Global) Static memory positions.
            Keep a map of all accessible memory positions to their source statements
            per function. After that, we traverse the CFG and link each pair of
            reads/writes together in the order of control-flow.
        - (Intra-functional) Indirect memory access
            TODO

        In this way it's an O(N) algorithm, where N is the number of all functions
        (with context sensitivity). The old worklist algorithm is exponential in the
        worst case.
        '''
        # Stack access
        # Unfortunately, we have no idea where a function is, and it's better
        # not to rely on function identification methods. So we just traverse
        # the CFG once, and maintain a map of scanned IRSBs so that we scan
        # each IRSB only once.
        scanned_runs = defaultdict(int)
        initial_node = self._cfg.get_any_node(self._entry_point)
        initial_irsb = self._cfg.irsb_from_node(initial_node)

        # Setup the stack range
        # TODO: We are assuming the stack is at most 8 KB
        stack_val = initial_irsb.initial_state.regs.sp
        stack_ubound = initial_irsb.initial_state.se.any_int(stack_val)
        stack_lbound = stack_ubound - 0x80000

        # We maintain a calling stack so that we can limit all analysis within
        # the range of a single function.
        # Without function analysis, our approach is quite simple: If there is
        # an SimExit whose jumpkind is 'Ijk_Call', then we create a new frame
        # in calling stack. A frame is popped out if there exists an SimExit
        # that has 'Ijk_Ret' as its jumpkind.
        initial_call_frame = StackFrame(initial_sp=stack_ubound)
        initial_wrapper = RunWrapper(initial_node,
                                     new_state=None,
                                     call_stack=[initial_call_frame])

        def _find_frame_by_addr(stack, addr):
            '''
            Try to find the right stack frame according to addr. All non-stack
            values go to the outermost stack frame.
            Returns the correct RunWrapper instance.
            '''
            if len(stack) == 0:
                raise Exception("Stack is empty")

            if not (addr >= stack_lbound and addr <= stack_ubound):
                return stack[0]

            for fr in reversed(stack):
                if fr.initial_sp is None:
                    return fr
                if addr < fr.initial_sp:
                    return fr
            return stack[0]

        def _find_top_frame(stack):
            if len(stack) == 0:
                raise Exception("Stack is empty")

            return stack[0]

        bbl_stmt_mem_map = defaultdict(lambda: defaultdict(set))
        returned_memory_addresses = set()

        # All pending SimRuns
        run_stack = [ initial_wrapper ]
        while len(run_stack) > 0:
            current_run_wrapper = run_stack.pop()

            node = current_run_wrapper.node
            run = self._cfg.irsb_from_node(node)
            l.debug("Picking %s... it has been analyzed %d times",
                    run, scanned_runs[node])
            if scanned_runs[node] >= MAX_ANALYZE_TIMES:
                continue
            else:
                scanned_runs[node] += 1
            # new_run = run.reanalyze(new_state=current_run_wrapper.new_state)
            # FIXME: Now we are always generating new SimRun to avoid issues in ARM mode
            if run.initial_state.arch.name == "ARM":
                new_run = self._project.factory.sim_run(self._project.exit_to(run.addr, state=current_run_wrapper.new_state))
            else:
                new_run = run.reanalyze(new_state=current_run_wrapper.new_state)
            l.debug("Scanning %s", new_run)

            reanalyze_successors_flag = current_run_wrapper.reanalyze_successors
            new_addr_written = False

            if isinstance(new_run, SimIRSB):
                old_irsb = run
                irsb = new_run

                # Simulate the execution of this IRSB.
                # For MemWriteRef, fill the addr_to_ref dict with every single
                # concretizable memory address, and just ignore symbolic ones
                # for now.
                # For MemReadRef, we get its related MemWriteRef from our dict
                stmts = irsb.statements
                for i in xrange(len(stmts)):
                    stmt = stmts[i]
                    actions = stmt.actions
                    if len(actions) == 0:
                        continue
                    # Obtain the real Ref of current statement
                    real_action = actions[-1]

                    if real_action.type == 'mem' and real_action.action == 'write':
                        addr = real_action.addr.ast
                        if not run.initial_state.se.symbolic(addr):
                            # It's not symbolic. Try to concretize it.
                            concrete_addr = run.initial_state.se.any_int(addr)
                            # Create the tuple of (simrun_addr, stmt_id)
                            tpl = (irsb.addr, i)
                            frame = _find_frame_by_addr(current_run_wrapper.call_stack, \
                                                     concrete_addr)
                            if concrete_addr in frame.addr_to_ref and \
                                frame.addr_to_ref[concrete_addr] == tpl:
                                pass
                            else:
                                frame.addr_to_ref[concrete_addr] = tpl
                                l.debug("Memory write to addr 0x%x, irsb %s, " + \
                                        "stmt id = %d", concrete_addr, irsb, i)
                                reanalyze_successors_flag = True
                                if concrete_addr not in bbl_stmt_mem_map[old_irsb][i]:
                                    bbl_stmt_mem_map[old_irsb][i].add(concrete_addr)
                                    new_addr_written = True
                        else:
                            # Add it to our symbolic memory operation list. We
                            # will process them later.
                            self._symbolic_mem_ops.add((old_irsb, real_action))
                            # FIXME
                            # This is an ugly fix, but let's stay with it for now.
                            # If this is a symbolic MemoryWrite, it's possible that
                            # all memory read later relies on it.
                            tpl = (irsb.addr, i)
                            top_frame = _find_top_frame(current_run_wrapper.call_stack)
                            top_frame.symbolic_writes.append(tpl)
                    for action in actions:
                        if action.type == 'mem' and action.action == 'read':
                            addr = action.addr.ast
                            # Relate it to all previous symbolic writes
                            top_frame = _find_top_frame(current_run_wrapper.call_stack)
                            for tpl in top_frame.symbolic_writes:
                                self._graph[irsb.addr][i].add(tpl)

                            if not run.initial_state.se.symbolic(addr):
                                # Not symbolic. Try to concretize it.
                                concrete_addr = run.initial_state.se.any_int(addr)
                                # Check if this address has been written before.
                                # Note: we should check every single call frame,
                                # from the latest to earliest, until we come
                                # across that address.
                                frame = _find_frame_by_addr(current_run_wrapper.call_stack, \
                                                         concrete_addr)
                                if concrete_addr in frame.addr_to_ref:
                                    # Luckily we found it!
                                    # Record it in our internal dict
                                    tpl = frame.addr_to_ref[concrete_addr]
                                    l.debug("Memory read to addr 0x%x, irsb %s, stmt id = %d, Source: <0x%x>[%d]", \
                                            concrete_addr, irsb, i, tpl[0], tpl[1])
                                    self._graph[irsb.addr][i].add(
                                        frame.addr_to_ref[concrete_addr])
                                    break
                                # TODO: What if we have never seen this address
                                # before? It might be an address whose value is
                                # initialized somewhere else, or an address that
                                # contains initialized value.
                            else:
                                self._symbolic_mem_ops.add((old_irsb, action))
            else:
                # SimProcedure
                old_sim_proc = run
                sim_proc = new_run

                actions = sim_proc.initial_state.log.actions
                for action in actions:
                    if action.type == 'mem' and action.action == 'write':
                        addr = action.addr.ast
                        if not run.initial_state.se.symbolic(addr):
                            # Record it
                            # Not symbolic. Try to concretize it.
                            concrete_addr = run.initial_state.se.any_int(addr)
                            # Create the tuple of (simrun_addr, stmt_id)
                            tpl = (sim_proc.addr, i)
                            frame = _find_frame_by_addr(current_run_wrapper.call_stack, \
                                                         concrete_addr)
                            if concrete_addr in frame.addr_to_ref and \
                                frame.addr_to_ref[concrete_addr] == tpl:
                                pass
                            else:
                                l.debug("Memory write to addr 0x%x, SimProc %s", \
                                        concrete_addr, sim_proc)
                                frame.addr_to_ref[concrete_addr] = tpl
                                reanalyze_successors_flag = True
                                if concrete_addr not in bbl_stmt_mem_map[old_sim_proc][i]:
                                    bbl_stmt_mem_map[old_sim_proc][i].add(concrete_addr)
                                    # new_addr_written = True
                        else:
                            self._symbolic_mem_ops.add((old_sim_proc, action))
                            # FIXME
                            # This is an ugly fix, but let's stay with it for now.
                            # If this is a symbolic MemoryWrite, it's possbiel that
                            # all memory read later relies on it.
                            tpl = (sim_proc.addr, i)
                            top_frame = _find_top_frame(current_run_wrapper.call_stack)
                            top_frame.symbolic_writes.append(tpl)
                    elif action.type == 'mem' and action.action == 'read':
                        addr = action.addr.ast
                        # Relate it to all previous symbolic writes
                        top_frame = _find_top_frame(current_run_wrapper.call_stack)
                        for tpl in top_frame.symbolic_writes:
                            self._graph[sim_proc.addr][i].add(tpl)

                        if not run.initial_state.se.symbolic(addr):
                            # Not symbolic. Try to concretize it.
                            concrete_addr = run.initial_state.se.any_int(addr)
                            frame = _find_frame_by_addr(current_run_wrapper.call_stack, \
                                                          concrete_addr)
                            if concrete_addr in frame.addr_to_ref:
                                # Luckily we found it!
                                # Record it in our internal dict
                                tpl = frame.addr_to_ref[concrete_addr]
                                l.debug("Memory read to addr 0x%x, SimProc %s, Source: <0x%x>[%d]", \
                                        concrete_addr, sim_proc, tpl[0], tpl[1])
                                self._graph[old_sim_proc.addr][-1].add(
                                    frame.addr_to_ref[concrete_addr])
                            else:
                                l.debug("Memory read to addr 0x%x, SimProc %s, no source available", \
                                        concrete_addr, sim_proc)
                            # TODO: what if we didn't see that address before?
                        else:
                            self._symbolic_mem_ops.add((old_sim_proc, action))

            # Expand the current SimRun
            all_successors = self._cfg.get_successors(node)
            real_successors = new_run.successors

            succ_targets = set()
            for successor in all_successors:
                succ_addr = successor.addr
                if succ_addr in succ_targets:
                    continue
                succ_targets.add(succ_addr)
                # Ideally we shouldn't see any new successors here
                new_successors = [suc for suc in real_successors if suc.se.exactly_n_int(suc.ip, 1)[0] == succ_addr]
                if len(new_successors) > 0:
                    new_state = new_successors[0]
                else:
                    l.warning("Run %s. Cannot find requesting target 0x%x", run, succ_addr)
                    new_state = self._cfg.irsb_from_node(successor).initial_state

                new_call_stack = copy.deepcopy(current_run_wrapper.call_stack) # Make a copy

                is_ret = False
                if new_run.successors:
                    if new_run.successors[0].scratch.jumpkind == "Ijk_Call":
                        # Create a new function frame
                        new_sp = new_state.regs.sp
                        new_sp_concrete = new_state.se.any_int(new_sp)
                        new_stack_frame = StackFrame(initial_sp=new_sp_concrete)
                        new_call_stack.append(new_stack_frame)
                    elif new_run.successors[0].scratch.jumpkind == "Ijk_Ret":
                        is_ret = True
                        if len(new_call_stack) > 1:
                            # Pop out the latest function frame
                            new_call_stack.pop()
                        else:
                            # We are returning from somewhere, but the stack is
                            # already empty.
                            # Something must have went wrong.
                            l.warning("Stack is already empty before popping things out")
                    else:
                        # Do nothing :)
                        pass

                # TODO: This is an ugly fix!
                # If this SimExit is a ret and it's returning an address, we continue the execution anyway
                '''
                clear_successors_ctr = False
                if is_ret and len(new_successors) == 1:
                    # Check if it's returning an address
                    # FIXME: This is for ARM32!
                    ret_reg_offset = 0 * 4 + 8
                    ret_value = new_successors[0].se.any_int(new_successors[0].registers.load(ret_reg_offset))
                    if ret_value not in returned_memory_addresses:
                        returned_memory_addresses.add(ret_value)
                        if (0xffffff00 - ret_value < 100000 and ret_value < 0xffffff00) or \
                                abs(ret_value - 0xc0000000) < 16000:\
                            # Iteratively remove all its successors from scanned_runs
                            l.debug("%s returns an memory address 0x%x. Remove all its successors.", run, ret_value)
                            clear_successors_ctr = True

                if run.addr == 0x40906a70 and new_addr_written:
                    l.debug("%s writes at a new address. Remove all its successors.", run)
                    clear_successors_ctr = True

                if clear_successors_ctr:
                    for k, v in self._cfg.get_all_successors(run).items():
                        for node in v:
                            if node != run:
                                scanned_runs[node] = 0
                '''

                if successor in scanned_runs:
                    if not (reanalyze_successors_flag and scanned_runs[successor] < MAX_ANALYZE_TIMES):
                        l.debug("Skipping %s, reanalyze_successors_flag = %d, scan times = %d", successor, reanalyze_successors_flag, scanned_runs[successor])
                        continue

                continue_flag = False
                for s in run_stack:
                    if s.node == successor:
                        continue_flag = True
                        break
                if continue_flag:
                    continue

                wrapper = RunWrapper(successor,
                        new_state=new_state,
                        call_stack=new_call_stack,
                        reanalyze_successors=reanalyze_successors_flag)
                run_stack.append(wrapper)
                l.debug("Appending successor %s.", successor)

        #self._solve_symbolic_mem_operations()

class StackFrame(object):
    def __init__(self, initial_sp, addr_to_ref=None):
        self.initial_sp = initial_sp
        if addr_to_ref is None:
            self.addr_to_ref = {}
        else:
            self.addr_to_ref = addr_to_ref
        self.symbolic_writes = []

class RunWrapper(object):
    '''
We keep an RunWrapper object for each function (with context sensitivity, it's
an AddrToRefContainer object for each [function, context] pair). It contains a
list of all runs inside this function, a dict addr_to_ref storing all
references between addresses and a [simrun_addr, stmt_id] pair, and a calling
stack.
    '''
# TODO: We might want to change the calling stack into a branching list with
# CoW supported.
    def __init__(self, node, new_state, call_stack=None, reanalyze_successors=False):
        self.node = node
        self.new_state = new_state

        if call_stack is None:
            self.call_stack = []
        else:
            # We DO NOT make a copy of the provided stack object
            self.call_stack = call_stack

        self.reanalyze_successors = reanalyze_successors



from collections import defaultdict
from itertools import ifilter
import logging

import networkx

import simuvex

from ..annocfg import AnnotatedCFG
from ..analysis import Analysis
from ..errors import AngrBackwardSlicingError

l = logging.getLogger(name="angr.analyses.backward_slicing")

class WorkList(object):
    def __init__(self):
        self._worklist = set()

    def add(self, taint_source):
        # Scan the current list for existing TaintSource object of the same run. If
        # there exists any, combine them.
        existing_items = filter(lambda r : r.run == taint_source.run, self._worklist)
        if len(existing_items) == 0:
            self._worklist.add(taint_source)
        elif len(existing_items) == 1:
            old_ts = existing_items[0]
            self._worklist.remove(old_ts)
            taint_source.data_taints |= old_ts.data_taints
            taint_source.reg_taints |= old_ts.reg_taints
            taint_source.tmp_taints |= old_ts.tmp_taints
            # Write our combined taint source to the worklist
            self._worklist.add(taint_source)
        else:
            raise Exception("Something is wrong!")

    def pop(self):
        if len(self._worklist) > 0:
            return self._worklist.pop()
        return None

    def size(self):
        return len(self._worklist)

    def items(self):
        for ts in self._worklist:
            yield ts

class TaintSet(object):
    def __init__(self, run, stmt_id, data_taints, reg_taints, tmp_taints, taints_waitinglist=None, kids=None,
                 parent=None):
        """
        Describes a set of all taints at the end of a SimRun. TaintSet instances are put into a worklist in our
        taint-tracking algorithm.

        :param run: The related SimRun
        :param stmt_id: ID of the he last statement that we care about. -1 stands for the very last statement of a basic
                        block.
        :param data_taints: A list of memory taints
        :param reg_taints: A list of register taints
        :param tmp_taints: A list of temporary variable taints
        :param taints_waitinglist:
        :param kids:
        :param parent:
        """
        kids = kids or [ ]
        taints_waitinglist = taints_waitinglist or {}

        self.run = run
        self.stmt_id = stmt_id
        self.data_taints = data_taints
        self.reg_taints = reg_taints
        self.tmp_taints = tmp_taints
        self.taints_waitinglist = taints_waitinglist
        self.kids = kids
        self.parent = parent

    def __repr__(self):
        return "<TaintSet mem:%s, reg:%s, tmp:%s>" % (self.data_taints, self.reg_taints, self.tmp_taints)

class Taint(object):
    def __init__(self, type, addr=None, stmt_id=None, reg=None, tmp=None, mem_addr=None, value=None):
        """
        Describes a taint.

        :param type: The type can be one of the following values: reg, tmp, mem, or const
        :param addr: Address where this taint is created
        :param stmt_id: ID of the statement where this taint is created
        :param reg: Register offset
        :param tmp: ID of the temporary variable
        :param mem_addr: Address in memory
        :param value: Other values. e.g. value of the constant
        """

        if type not in ('reg', 'tmp', 'mem', 'const', 'exit'):
            raise AngrBackwardSlicingError('Unsupported taint type "%s"' % type)

        self.type = type
        self.addr = addr
        self.stmt_id = stmt_id

        # TODO: More sanity checks
        self.reg = reg
        self.tmp = tmp
        self.mem_addr = mem_addr
        self.value = value

    def __eq__(self, other):
        if self.type == 'exit':
            return self.addr == other.addr and self.stmt_id == other.stmt_id
        if self.type == 'mem':
            mem_addr_equal = hash(self.mem_addr.model) == hash(other.mem_addr.model) if self.mem_addr.symbolic else \
                self.mem_addr.model == other.mem_addr.model

            return mem_addr_equal and self.type == other.type and self.addr == other.addr \
                    and self.stmt_id == other.stmt_id
        else:
            return self.type == other.type and self.addr == other.addr \
                   and self.stmt_id == other.stmt_id and self._data == other._data

    def __hash__(self):
        if self.type == 'exit':
            return hash((self.addr, self.stmt_id))
        if self.type == 'mem' and self.mem_addr.symbolic:
            h = hash(self.type + "_" + str(hash(self.mem_addr.model)) + "_" + str(self.addr) + "_" + str(self.stmt_id))
        else:
            h = hash(self.type + "_" + str(self._data) + "_" + str(self.addr) + "_" + str(self.stmt_id))
        return h

    def copy(self):
        return Taint(self.type,
                     addr=self.addr,
                     stmt_id=self.stmt_id,
                     reg=self.reg,
                     tmp=self.tmp,
                     mem_addr=self.mem_addr,
                     value=self.value)

    @property
    def _data(self):
        if self.type == 'exit':
            data = None
        elif self.type == 'reg':
            data = self.reg
        elif self.type == 'tmp':
            data = self.tmp
        elif self.type == 'mem':
            data = self.mem_addr
        else:
            data = self.value

        return data

    def __repr__(self):
        if self.type == 'exit':
            s = "Taint<%s>(%s:%s)" % (self.type, hex(self.addr), self.stmt_id)
        else:
            s = "Taint<%s %s>(%s:%s)" % (self.type, self._data, hex(self.addr) if self.addr else self.addr, self.stmt_id)
        return s

class DataTaint(object):
    def __init__(self, simrun_addr, stmt_idx, address=None, bits=None):
        """
        Descriptor for tainted memory addresses and/or the statement that introduces this taint. It is put into
        data_taint_set during the worklist
        """

        self.simrun_addr = simrun_addr
        self.stmt_idx = stmt_idx
        self.address = address
        self.bits = bits

    @property
    def has_address(self):
        """
        Is this data taint descriptor related to some address?

        :return: True/False
        """
        return (self.address is not None)

    def __repr__(self):
        if not self.address:
            return "<DataTaint stmt %s:%d>" % (hex(self.simrun_addr), self.stmt_idx)
        else:
            return "<DataTaint stmt %s:%d addr %s[%d]>" % (
                hex(self.simrun_addr),
                self.stmt_idx,
                self.address,
                self.bits.ast
            )

    def __eq__(self, other):
        if not self.has_address:
            return self.simrun_addr == other.simrun_addr and self.stmt_idx == other.stmt_idx

        else:
            return self.address == other.address and self.bits == other.bits

    def __hash__(self):
        if not self.has_address:
            return hash((self.simrun_addr, self.stmt_idx))

        else:
            return hash((self.address, self.bits))

class BackwardSlice(Analysis):

    def __init__(self, cfg, cdg, ddg, irsb, stmt_id,
                 control_flow_slice=False,
                 no_construct=False):
        '''

        :param cfg:
        :param cdg:
        :param ddg:
        :param irsb:
        :param stmt_id:
        :param control_flow_slice:
        :param no_construct:            Only used for testing and debugging to easily create a BackwardSlice object
        :return:
        '''
        self._project = self._p
        self._cfg = cfg
        self._cdg = cdg
        self._ddg = ddg

        self._target_irsb = irsb
        self._target_stmt_idx = stmt_id

        # Save a list of taints to beginwwith at the beginning of each SimRun
        self.initial_taints_per_run = None
        self.runs_in_slice = None
        self.run_statements = None

        if not no_construct:
            self._construct(irsb, stmt_id, control_flow_slice=control_flow_slice)

    def annotated_cfg(self, start_point=None):
        '''
        Returns an AnnotatedCFG based on slicing result.
        '''

        target_irsb_addr = self._target_irsb.addr
        target_stmt = self._target_stmt_idx
        start_point = start_point if start_point is not None else self._p.entry

        l.debug("Initializing AnnoCFG...")
        target_irsb = self._cfg.get_any_node(target_irsb_addr)
        anno_cfg = AnnotatedCFG(self._project, self._cfg, target_irsb_addr)
        if target_stmt is not -1:
            anno_cfg.set_last_stmt(target_irsb, target_stmt)

        #start_point_addr = 0
        #successors = [self._cfg.get_any_node(start_point)]
        #processed_successors = set()
        #while len(successors) > 0:

        for n in self._cfg.graph.nodes():
            run = n

            if run in self.run_statements:
                if self.run_statements[run] is True:
                    anno_cfg.add_simrun_to_whitelist(run)
                else:
                    anno_cfg.add_statements_to_whitelist(run, self.run_statements[run])

        for src, dst in self._cfg.graph.edges():

            run = src

            if dst in self.run_statements and src in self.run_statements:
                anno_cfg.add_exit_to_whitelist(run, dst)

            # TODO: expose here, maybe?
            #anno_cfg.set_path_merge_points(self._path_merge_points)

        return anno_cfg

    def is_taint_ip_related(self, simrun_addr, stmt_idx, taint_type, simrun_whitelist=None):
        """
        Query in taint graph to check if a specific taint will taint the IP in the future or not.
        The taint is specified with the tuple (simrun_addr, stmt_idx, taint_type).

        :param simrun_addr: Address of the SimRun
        :param stmt_idx: Statement ID
        :param taint_type: Type of the taint, might be one of the following: 'reg', 'tmp', 'mem'
        :param simrun_whitelist: A list of SimRun addresses that are whitelisted, i.e. the tainted exit will be ignored
                                if it is in those SimRuns
        :return: True/False
        """

        if simrun_whitelist is None:
            simrun_whitelist = set()
        if type(simrun_whitelist) is not set:
            simrun_whitelist = set(simrun_whitelist)

        # Find the specific taint in our graph
        taint = None
        for n in self.taint_graph.nodes():
            if n.type == taint_type and n.addr == simrun_addr and n.stmt_id == stmt_idx:
                taint = n
                break

        if taint is None:
            raise AngrBackwardSlicingError('The specific taint is not found')

        bfs_tree = networkx.bfs_tree(self.taint_graph, taint)

        # A node is tainting the IP if one of the following criteria holds:
        # - a descendant tmp variable is used as a default exit or a conditional exit of its corresponding SimRun
        # - a descendant register is the IP itself

        for descendant in bfs_tree.nodes():
            if descendant.type == 'exit':
                if descendant.addr not in simrun_whitelist:
                    return True
            elif descendant.type == 'reg' and descendant.reg == self._p.arch.ip_offset:
                return True

        return False

    def is_taint_impacting_stack_pointers(self, simrun_addr, stmt_idx, taint_type, simrun_whitelist=None):
        """
        Query in taint graph to check if a specific taint will taint the stack pointer in the future or not.
        The taint is specified with the tuple (simrun_addr, stmt_idx, taint_type).

        :param simrun_addr: Address of the SimRun
        :param stmt_idx: Statement ID
        :param taint_type: Type of the taint, might be one of the following: 'reg', 'tmp', 'mem'
        :param simrun_whitelist: A list of SimRun addresses that are whitelisted
        :return: True/False
        """

        if simrun_whitelist is None:
            simrun_whitelist = set()
        if type(simrun_whitelist) is not set:
            simrun_whitelist = set(simrun_whitelist)

        # Find the specific taint in our graph
        taint = None
        for n in self.taint_graph.nodes():
            if n.type == taint_type and n.addr == simrun_addr and n.stmt_id == stmt_idx:
                taint = n
                break

        if taint is None:
            raise AngrBackwardSlicingError('The specific taint is not found')

        bfs_tree = networkx.bfs_tree(self.taint_graph, taint)

        # A node is tainting the stack pointer if one of the following criteria holds:
        # - a descendant register is the sp/bp itself

        for descendant in bfs_tree.nodes():
            if descendant.type == 'reg' and (
                        descendant.reg in (self._p.arch.sp_offset, self._p.arch.bp_offset)
            ):
                return True

        return False

    #
    # Private methods
    #

    def _construct(self, sim_run, stmt_id, control_flow_slice=False):
        """
        Construct a dependency graph based on given parameters.

        :param sim_run: The SimRun instance where the backward slice should start from
        :param stmt_id: The statement ID where the backward slice starts from in the target SimRun instance. -1 for
                        starting from the very last statement
        :param control_flow_slice: Is the backward slicing only depends on CFG or not
        :return: None
        """

        if control_flow_slice:
            self._construct_control_flow_slice(sim_run)

        else:
            self._construct_default(sim_run, stmt_id)

    def _construct_control_flow_slice(self, irsb):
        """
        Build a slice of the program without considering the effect of data dependencies.
        This ia an incorrect hack, but it should work fine with small programs.

        :param irsb: The target IRSB. You probably wanna get it from the CFG somehow. It
                    must exist in the CFG.
        :return: None
        """

        if self._cfg is None:
            l.error('Please build CFG first.')

        cfg = self._cfg.graph
        if irsb not in cfg:
            l.error('SimRun instance %s is not in the CFG.', irsb)

        reversed_cfg = networkx.DiGraph()
        # Reverse the graph
        for s, d in cfg.edges():
            reversed_cfg.add_edge(d, s)

        # Traverse forward in the reversed graph
        stack = [ ]
        stack.append(irsb)

        self.runs_in_slice = networkx.DiGraph()

        self.run_statements = { }
        while stack:
            # Pop one out
            block = stack.pop()
            if block not in self.run_statements:
                self.run_statements[block] = True
                # Get all successors of that block
                successors = reversed_cfg.successors(block)
                for succ in successors:
                    stack.append(succ)
                    self.runs_in_slice.add_edge(succ, block)

    def _construct_default(self, cfg_node, stmt_id):
        """
        Create a backward slice from a specific statement in a specific sim_run. It's based on a worklist algorithm.

        :param cfg_node: The CFGNode instance where the backward slice starts. It must be in the CFG
        :param stmt_id: ID of the target statement where the backward slice starts
        :return: None
        """

        self.taint_graph = networkx.DiGraph()

        if cfg_node not in self._cfg.graph:
            raise AngrBackwardSlicingError('Target SimRun (%s) is not in the CFG.', cfg_node)

        if stmt_id == -1:
            # Its jump target is tainted
            data_reg_deps = set()
            data_tmp_deps = set()

            # The tmp variable that irsb.next relies on will be tainted later

        else:
            path = self._project.factory.path(irsb.initial_state)
            s = path.state
            actions = filter(lambda r: r.type == 'reg' and r.action == 'write' and r.stmt_idx == stmt_id,
                          s.log.actions)
            if len(actions) != 1:
                raise Exception("Invalid references. len(actions) == %d." % len(actions))
            data_reg_deps = set(actions[0].data.reg_deps)
            data_tmp_deps = set(actions[0].data.tmp_deps)

        # TODO: Make it more elegant
        data_deps = set()
        # TODO: What's the data dependency here?

        start = TaintSet(cfg_node, stmt_id, data_deps, data_reg_deps, data_tmp_deps, taints_waitinglist={ }, kids=[ ])

        worklist = WorkList()
        worklist.add(start)
        processed_ts = set()
        self.initial_taints_per_run = defaultdict(list)
        self.runs_in_slice = networkx.DiGraph()
        # We are using a list here, and later on we reconstruct lists and write it to
        # self.run_statements
        self.run_statements = defaultdict(set)

        while worklist.size() > 0:

            ts = worklist.pop()

            # Initialize a temporary worklist
            temp_worklist = WorkList()

            # The current SimRun we are dealing with in this iteration
            try:
                current_run = self._p.factory.sim_run(ts.run.input_state)
            except simuvex.SimIRSBError:
                continue

            # Make a copy of all taint sets
            data_taint_set = ts.data_taints.copy()
            reg_taint_set = ts.reg_taints.copy()
            tmp_taint_set = ts.tmp_taints.copy()

            # A waiting list of all taints that are yet to be added onto the taint graph
            taints_waitinglist = ts.taints_waitinglist

            l.debug("Pick a new SimRun %s", current_run)

            if isinstance(current_run, simuvex.SimIRSB):
                current_run_type = 'irsb'

            elif isinstance(current_run, simuvex.SimProcedure):
                current_run_type = 'simproc'


            else:
                raise Exception("Unsupported SimRun type %s" % type(ts.run))

            state = None
            # We pick the state that has the most SimActions
            # TODO: Maybe we should always pick the one that is the default exit?
            max_count = 0
            successors = current_run.successors if current_run.successors else current_run.unsat_successors
            for s in successors:
                actions = list(s.log.actions)
                if len(actions) > max_count:
                    max_count = len(actions)
                    state = s

            if state is None:
                continue

            # We always taint the IP, otherwise Slicecutor cannot execute the generated slice
            reg_taint_set.add(self._project.arch.ip_offset)

            # Traverse the the current SimRun, and taint everything related

            # Now we don't have to taint the default exit anymore since Yan is kind enough to provide a SimExitAction!
            """
            # Taint the default exit first
            for action in current_run.next_expr.actions:
                if action.type == "tmp" and action.action == "read":
                    tmp_taint_set.add(action.tmp)
            """

            # However, we have to taint all exits other than the default one, otherwise the corresponding exit will be
            # missing in the final slice
            if current_run_type == 'irsb':
                all_exit_stmts = [ (i, stmt) for (i, stmt) in enumerate(current_run.irsb.statements)
                                   if isinstance(stmt, pyvex.IRStmt.Exit) ]
            else:
                all_exit_stmts = [ ]

            for i, exit_stmt in all_exit_stmts:
                tmp_taint_set.add(exit_stmt.guard.tmp)
                self.run_statements[ts.run].add(i)

            # We also taint the stack pointer, so we could keep the stack balanced
            reg_taint_set.add(self._project.arch.sp_offset)

            stmt_start_id = ts.stmt_id
            if stmt_start_id == -1:
                actions = reversed(list(state.log.actions))
            else:
                actions = reversed([a for a in state.log.actions if a.stmt_idx <= stmt_start_id])

            # Taint the tmp variable that irsb.next relies on
            if stmt_start_id == -1:
                # Get the basic block
                # TODO: Make sure we are working on an IRSB, not a SimProcedure
                # TODO: Get opt_level from state options
                pyvex_irsb = self._p.factory.block(addr=current_run.addr, opt_level=1)
                irsb_next = pyvex_irsb.vex.next

                if type(irsb_next) is pyvex.IRExpr.RdTmp:
                    tmp_taint_set.add(irsb_next.tmp)

            # Group those actions from the same statement together
            statement_actions = defaultdict(list)
            for action in actions:
                statement_actions[action.stmt_idx].append(action)

            all_stmt_idx = sorted(statement_actions.keys(), reverse=True)

            for stmt_idx in all_stmt_idx:

                # For every statement, the mem read action is taken and its corresponding addresses are labeled as
                # tainted if its tmp_deps/reg_deps are tainted *locally (e.g. in current statement)*. In this way,
                # we don't lose the link between a tmp write and a mem read.
                #
                # For example, the following statement
                #   t16 = LDle:I32(t14)
                # creates three SimActions:
                # - a reg read, reading from t14
                # - a mem read, reading from *t14
                # - a tmp write, writing to t16
                # The connection between the tmp write and the mem read is currently missing in SimActions. The only
                # way to handle it correctly is to taint everything in this statement immediately after we see a
                # tmp write to a tainted temp variable occurs.

                all_actions = statement_actions[stmt_idx]

                # All tainted registers and temp variables in current statement
                local_reg_taint_records = set()
                local_tmp_taint_records = set()

                for action in all_actions:

                    if not hasattr(action, 'action'):
                        # SimExitAction
                        handler_name = '_handle_%s' % action.type

                    else:
                        handler_name = "_handle_%s_%s" % (action.type, action.action)

                    if hasattr(self, handler_name):
                        action_taken = False

                        if action.type == 'mem' and action.action == 'read':
                            # Whether taking the mem read or not depends on whether it relies on registers/tmps
                            # that are tainted locally.
                            if (action.tmp_deps.issubset(local_tmp_taint_records) and
                                    action.reg_deps.issubset(local_reg_taint_records)):
                                handler = getattr(self, handler_name)
                                action_taken = handler(action, tmp_taint_set, reg_taint_set, data_taint_set, taints_waitinglist,
                                                        temp_worklist, ts.run, current_run, action.stmt_idx)

                        else:
                            handler = getattr(self, handler_name)
                            action_taken = handler(action, tmp_taint_set, reg_taint_set, data_taint_set, taints_waitinglist,
                                                temp_worklist, ts.run, current_run, action.stmt_idx)

                        if action_taken:
                            for r in action.reg_deps:
                                local_reg_taint_records.add(r)
                            for t in action.tmp_deps:
                                local_tmp_taint_records.add(t)

            # We create the TaintSource object using the old taints from ts, not the new ones
            # (tmp_taint_set, reg_taint_set, and data_taint_set). Then TS object is put into
            # processed_ts.
            new_taint_set = TaintSet(ts.run, -1, ts.data_taints, ts.reg_taints, ts.tmp_taints,
                                      taints_waitinglist=taints_waitinglist, kids=ts.kids)
            processed_ts.add(new_taint_set)

            # Filter all tainted registers
            has_reg_taints = False
            if len(reg_taint_set - ignored_registers[self._p.arch.name]) > 0:
                has_reg_taints = True

            # In the end, add this taint set to our list
            initial_taint_set = TaintSet(ts.run, -1, data_taint_set.copy(), reg_taint_set.copy(), tmp_taint_set.copy())
            self.initial_taints_per_run[ts.run].append(initial_taint_set)

            if len(data_taint_set) > 0 or has_reg_taints:
                # Get its predecessors from our CFG
                predecessors = self._cfg.get_predecessors(ts.run)
                for p in predecessors:
                    new_reg_taint_set = reg_taint_set.copy()
                    new_data_taint_set = data_taint_set.copy()
                    new_tmp_taint_set = tmp_taint_set.copy()
                    kids_set = set()

                    if self._p.is_hooked(p.addr):
                        irsb_ = self._cfg.irsb_from_node(p)
                        # If it's not a boring exit from its predecessor, we shall
                        # search for the last branching, and taint the temp variable
                        # there.
                        flat_successors = irsb_.flat_successors
                        # Remove the simulated return exit
                        # YAN: commented this out because the pseudo-rets no longer show as reachable, and flat_exits takes reachable by default
                        #if len(flat_exits) > 0 and \
                        #       flat_exits[0].jumpkind == "Ijk_Call":
                        #   assert flat_exits[-1].jumpkind == "Ijk_Ret"
                        #   del flat_exits[-1]
                        if len(flat_successors) > 1:
                            successors = [suc for suc in flat_successors if
                                    not suc.se.symbolic(suc.ip) and
                                    suc.se.exactly_int(suc.ip, default=None) == ts.run.addr]
                            if len(successors) == 0 or not self._p.factory.sim_run(successors[0]).default_exit:
                                # It might be 0 sometimes...
                                # Search for the last branching exit, just like
                                #     if (t12) { PUT(184) = 0xBADF00D:I64; exit-Boring }
                                # , and then taint the temp variable inside if predicate
                                cmp_stmt_id, cmp_tmp_id = self._last_branching_statement(p.statements)
                                if cmp_stmt_id is not None:
                                    new_tmp_taint_set.add(cmp_tmp_id)
                                    self.run_statements[p].add(cmp_stmt_id)

                    l.debug("%s Got new predecessor %s", ts.run, p)
                    new_ts = TaintSet(p, -1, new_data_taint_set, new_reg_taint_set, new_tmp_taint_set,
                                      taints_waitinglist=taints_waitinglist.copy(), kids=list(kids_set), parent=ts.run)
                    temp_worklist.add(new_ts)
                    # Add it to our map
                    self.runs_in_slice.add_edge(p, ts.run)

            # Let's also search for predecessors on control dependency graph
            cdg_predecessors = self._cdg.get_predecessors(ts.run) if self._cdg is not None else [ ]
            for p in cdg_predecessors:
                new_reg_taint_set = reg_taint_set.copy()
                new_data_taint_set = data_taint_set.copy()
                new_tmp_taint_set = set()
                kids_set = set()
                if self._p.is_hooked(p.addr):
                    # Search for the last branching exit, just like
                    #     if (t12) { PUT(184) = 0xBADF00D:I64; exit-Boring }
                    # , and then taint the temp variable inside if predicate
                    irsb_ = self._cfg.irsb_from_node(p.run)
                    cmp_stmt_id, cmp_tmp_id = self._last_branching_statement(irsb_.statements)
                    if cmp_stmt_id is not None:
                        new_tmp_taint_set.add(cmp_tmp_id)
                        self.run_statements[p].add(cmp_stmt_id)

                l.debug("%s Got new control-dependency predecessor %s", ts.run, p)
                new_ts = TaintSet(p, -1, new_data_taint_set, new_reg_taint_set, new_tmp_taint_set,
                                  taints_waitinglist=taints_waitinglist.copy(), kids=list(kids_set))
                temp_worklist.add(new_ts)
                # Add it to our map
                self.runs_in_slice.add_edge(p, ts.run)

            # Merge temp worklist with the real worklist
            for taint_set in temp_worklist.items():
                # Merge it with existing items in processed_ts
                existing_tses = filter(lambda r : r.run == taint_set.run, processed_ts) #pylint:disable=W0640
                if len(existing_tses) > 0:
                    existing_ts = existing_tses[0]
                    if (existing_ts.reg_taints >= taint_set.reg_taints and
                             # existing_ts.data_taints >= taint_set.data_taints and
                            existing_ts.tmp_taints >= taint_set.tmp_taints):
                        l.debug("Ignore predecessor %s", taint_set.run)
                        continue
                    else:
                        # Remove the existing one
                        processed_ts.remove(existing_ts)
                        # Merge the old taint sets into new taint sets
                        taint_set.reg_taints |= existing_ts.reg_taints
                        taint_set.data_taints |= existing_ts.data_taints
                        taint_set.tmp_taints |= existing_ts.tmp_taints

                l.debug("%s added to real worklist.", taint_set.run)
                # Add it into the real worklist
                worklist.add(taint_set)

        # Reconstruct the run_statements
        _run_statements = self.run_statements
        self.run_statements = defaultdict(list)
        for cfg_node, s in _run_statements.items():
            self.run_statements[cfg_node] = list(s)

    #
    # All handlers
    #

    def _handle_exit(self, action, tmp_taint_set, reg_taint_set, data_taint_set,
                          taints_waitinglist, temp_worklist, cfg_node, current_run, stmt_idx):
        """
        Slicing handler for register writes

        :param action: The SimAction instance
        :param tmp_taint_set: A set of all tainted temp variable indices
        :param reg_taint_set: A set of all tainted register offsets
        :param data_taint_set: A set of all tainted memory addresses
        :param taints_waitinglist:
        ;param temp_worklist: The temporary worklist
        :param cfg_node: The CFG node
        :param current_run: Current SimRun instance
        :param stmt_idx: Statement ID. -1 for SimProcedures
        :return: True/False, indicating whether the corresponding statement is taken or not
        """

        # The exit is always tainted

        for tmp_dep in action.tmp_deps:
            tmp_taint_set.add(tmp_dep)
            # Create a new tmp taint, and put it into the waiting list
            taint = Taint('tmp', tmp=tmp_dep)
            taints_waitinglist[taint] = Taint('exit', addr=current_run.addr, stmt_id=stmt_idx)

        return True

    def _handle_reg_write(self, action, tmp_taint_set, reg_taint_set, data_taint_set,
                          taints_waitinglist, temp_worklist, cfg_node, current_run, stmt_idx):
        """
        Slicing handler for register writes

        :param action: The SimAction instance
        :param tmp_taint_set: A set of all tainted temp variable indices
        :param reg_taint_set: A set of all tainted register offsets
        :param data_taint_set: A set of all tainted memory addresses
        :param taints_waitinglist:
        ;param temp_worklist: The temporary worklist
        :param cfg_node: The CFG node
        :param current_run: Current SimRun instance
        :param stmt_idx: Statement ID. -1 for SimProcedures
        :return: True/False, indicating whether the corresponding statement is taken or not
        """

        reg_offset = action.offset

        if reg_offset not in reg_taint_set:
            return False

        else:
            self.run_statements[cfg_node].add(stmt_idx)

            # Add all taint links onto our taint graph
            waiting_taints = [ (s, d) for s, d in taints_waitinglist.iteritems()
                         if s.type == 'reg' and s.reg == reg_offset ]
            for src, dst in waiting_taints:
                del taints_waitinglist[src]

                src = src.copy()
                src.addr = current_run.addr
                src.stmt_id = stmt_idx
                self.taint_graph.add_edge(src, dst)

            if action.offset != self._project.arch.ip_offset:
                # Remove this taint
                reg_taint_set.remove(reg_offset)

            # Taint all its dependencies
            for reg_dep in action.data.reg_deps:
                reg_taint_set.add(reg_dep)
                # Create a new reg taint, and put it into the waiting list
                taint = Taint('reg', reg=reg_dep)
                taints_waitinglist[taint] = Taint('reg', addr=current_run.addr, stmt_id=stmt_idx, reg=reg_offset)

            for tmp_dep in action.data.tmp_deps:
                tmp_taint_set.add(tmp_dep)
                # Create a new tmp taint, and put it into the waiting list
                taint = Taint('tmp', tmp=tmp_dep)
                taints_waitinglist[taint] = Taint('reg', addr=current_run.addr, stmt_id=stmt_idx, reg=reg_offset)

            return True

    def _handle_tmp_write(self, action, tmp_taint_set, reg_taint_set, data_taint_set,
                          taints_waitinglist, temp_worklist, cfg_node, current_run, stmt_idx):
        """
        Slicing handler for temporary variable assignments

        :param action: The SimAction instance
        :param tmp_taint_set: A set of all tainted temp variable indices
        :param reg_taint_set: A set of all tainted register offsets
        :param data_taint_set: A set of all tainted memory addresses
        :param taints_waitinglist:
        ;param temp_worklist: The temporary worklist
        :param cfg_node: The CFG node
        :param current_run: Current SimRun instance
        :param stmt_idx: Statement ID. -1 for SimProcedures
        :return: True/False, indicating whether the corresponding statement is taken or not
        """

        tmp = action.tmp

        if tmp not in tmp_taint_set:
            return False

        else:
            self.run_statements[cfg_node].add(stmt_idx)

            # Retrieve all corresponding taints in the waiting list
            waiting_taints = [ (s, d) for s, d in taints_waitinglist.iteritems()
                               if s.type == 'tmp' and s.tmp == tmp ]
            for src, dst in waiting_taints:
                del taints_waitinglist[src]

                src = src.copy()
                src.addr = current_run.addr
                src.stmt_id = stmt_idx
                self.taint_graph.add_edge(src, dst)

            # Remove this taint
            tmp_taint_set.remove(tmp)

            # Taint all its dependencies
            for reg_dep in action.data.reg_deps:
                reg_taint_set.add(reg_dep)
                # Create a new reg taint, and put it onto the waiting list
                taint = Taint('reg', reg=reg_dep)
                taints_waitinglist[taint] = Taint('tmp', addr=current_run.addr, stmt_id=stmt_idx, tmp=tmp)

            for tmp_dep in action.data.tmp_deps:
                tmp_taint_set.add(tmp_dep)
                # Create a new tmp taint, and put it onto the waiting list
                taint = Taint('tmp', tmp=tmp_dep)
                taints_waitinglist[taint] = Taint('tmp', addr=current_run.addr, stmt_id=stmt_idx, tmp=tmp)

            # It might have a memory dependency
            # Take a look at the corresponding tmp
            stmt = current_run.statements[stmt_idx]
            mem_actions = [ a for a in stmt.actions if a.type == 'mem' and a.action == 'read' ]
            if mem_actions:
                mem_action = mem_actions[0]
                src = Taint(type="mem", addr=current_run.addr, stmt_id=stmt_idx, mem_addr=mem_action.addr.ast)
                dst = Taint(type="tmp", addr=current_run.addr, stmt_id=stmt_idx, tmp=tmp)
                self.taint_graph.add_edge(src, dst)

            return True

    def _handle_mem_read(self, action, tmp_taint_set, reg_taint_set, data_taint_set,
                         taints_waitinglist, temp_worklist, cfg_node, current_run, stmt_idx):
        """
        Slicing handler for memory reads

        :param action: The SimAction instance
        :param tmp_taint_set: A set of all tainted temp variable indices
        :param reg_taint_set: A set of all tainted register offsets
        :param data_taint_set: A set of all tainted memory addresses
        :param taints_waitinglist:
        ;param temp_worklist: The temporary worklist
        :param cfg_node: The CFG node
        :param current_run: Current SimRun instance
        :param stmt_idx: Statement ID. -1 for SimProcedures
        :return: True/False, indicating whether the corresponding statement is taken or not
        """

        # If this handler is called, this mem read action must be taken.

        if self._ddg is not None:

            # We rely on DDG if there is one available
            if (current_run.addr in self._ddg._graph and
                stmt_idx in self._ddg._graph[current_run.addr]):

                dependency_set = self._ddg._graph[current_run.addr][stmt_idx]

                for dependent_run_addr, dependent_stmt_id in dependency_set:
                    dependent_run = self._cfg.get_any_irsb(dependent_run_addr)
                    if isinstance(dependent_run, simuvex.SimIRSB):
                        # It's incorrect to do this:
                        # 'run_statements[dependent_run].add(dependent_stmt_id)'
                        # We should add a dependency to that SimRun object, and reanalyze
                        # it by putting it to our worklist once more

                        # Check if we need to reanalyze that block

                        data_taint = DataTaint(dependent_run.addr, dependent_stmt_id)

                        new_data_taint_set = set()
                        new_data_taint_set.add(data_taint)

                        dependent_runs = self._cfg.get_all_nodes(dependent_run_addr)
                        for d_run in dependent_runs:
                            new_ts = TaintSet(d_run, -1,
                                new_data_taint_set, set(), set())
                            temp_worklist.add(new_ts)
                            l.debug("%s added to temp worklist.", d_run)

                    else:
                        # A SimProcedure instance
                        data_taint = DataTaint(dependent_run.addr, -1)

                        new_data_taint_set = { data_taint }
                        dependent_runs = self._cfg.get_all_nodes(dependent_run_addr)
                        for d_run in dependent_runs:
                            new_ts = TaintSet(d_run, -1, new_data_taint_set,
                                            set(), set())
                            temp_worklist.add(new_ts)
                        l.debug("%s added to temp worklist.", dependent_run)

        else:
            # No DDG available. Let's log the address and size then

            data_taint = DataTaint(current_run.addr, stmt_idx, address=action.addr, bits=action.size)

            data_taint_set.add(data_taint)

            # Add all taint links onto our taint graph
            waiting_taints = [ (s, d) for s, d in taints_waitinglist.iteritems()
                         if s.type == 'mem' and (s.mem_addr == action.addr).model is True ] # TODO: mem_size
            for src, dst in waiting_taints:
                del taints_waitinglist[src]

                src = src.copy()
                src.addr = current_run.addr
                src.stmt_id = stmt_idx
                self.taint_graph.add_edge(src, dst)

            # Taint all its dependencies

            # for mem read, we should taint its source first
            taint = Taint('mem', mem_addr=action.addr.ast)
            taints_waitinglist[taint] = Taint('mem', addr=current_run.addr, stmt_id=stmt_idx, mem_addr=action.addr.ast) # TODO: mem_size

            for reg_dep in action.data.reg_deps:
                reg_taint_set.add(reg_dep)
                # Create a new reg taint, and put it into the waiting list
                taint = Taint('reg', reg=reg_dep)
                taints_waitinglist[taint] = Taint('mem', addr=current_run.addr, stmt_id=stmt_idx, mem_addr=action.addr.ast) # TODO: mem_size

            for tmp_dep in action.data.tmp_deps:
                tmp_taint_set.add(tmp_dep)
                # Create a new tmp taint, and put it into the waiting list
                taint = Taint('tmp', tmp=tmp_dep)
                taints_waitinglist[taint] = Taint('mem', addr=current_run.addr, stmt_id=stmt_idx, mem_addr=action.addr.ast) # TODO: mem_size

        return True

    def _handle_mem_write(self, action, tmp_taint_set, reg_taint_set, data_taint_set,
                          taints_waitinglist, temp_worklist, cfg_node, current_run, stmt_idx):
        """
        Slicing handler for memory writes

        :param action: The SimAction instance
        :param tmp_taint_set: A set of all tainted temp variable indices
        :param reg_taint_set: A set of all tainted register offsets
        :param data_taint_set: A set of all tainted memory addresses
        :param taints_waitinglist:
        ;param temp_worklist: The temporary worklist
        :param cfg_node: The CFG node
        :param current_run: Current SimRun instance
        :param stmt_idx: Statement ID. -1 for SimProcedures
        :return: True/False, indicating whether the corresponding statement is taken or not
        """

        if self._ddg is not None:
            # We can rely on the DDG
            data_taint = None
            for dt in data_taint_set:
                if dt.simrun_addr == current_run.addr and dt.stmt_idx == stmt_idx:
                    data_taint = dt
                    break

            if data_taint is not None:

                data_taint_set.remove(data_taint)

                self.run_statements[cfg_node].add(stmt_idx)

                for d in action.data.reg_deps:
                    reg_taint_set.add(d)
                for d in action.addr.reg_deps:
                    reg_taint_set.add(d)
                for d in action.data.tmp_deps:
                    tmp_taint_set.add(d)
                for d in action.addr.tmp_deps:
                    tmp_taint_set.add(d)
                # TODO: How do we handle other data dependencies here? Or if there is any?

                return True

            else:
                return False

        else:
            # DDG is not there...

            data_taint = None
            for dt in data_taint_set:
                if ((dt.address == action.addr).model is True # FIXME: This is ugly. claripy.is_true() is the way to go
                        and (dt.bits.ast == action.size.ast)):
                    data_taint = dt
                    break

            if data_taint is not None:

                data_taint_set.remove(data_taint)

                self.run_statements[cfg_node].add(stmt_idx)

                for d in action.data.reg_deps:
                    reg_taint_set.add(d)
                for d in action.addr.reg_deps:
                    reg_taint_set.add(d)
                for d in action.data.tmp_deps:
                    tmp_taint_set.add(d)
                for d in action.addr.tmp_deps:
                    tmp_taint_set.add(d)

                # Add all taint links onto our taint graph
                waiting_taints = [ (s, d) for s, d in taints_waitinglist.iteritems()
                             if s.type == 'mem' and (s.mem_addr == action.addr).model is True ] # TODO: mem_size
                for src, dst in waiting_taints:
                    del taints_waitinglist[src]

                    src = src.copy()
                    src.addr = current_run.addr
                    src.stmt_id = stmt_idx
                    self.taint_graph.add_edge(src, dst)

                # Taint all its dependencies
                for reg_dep in action.data.reg_deps:
                    reg_taint_set.add(reg_dep)
                    # Create a new reg taint, and put it into the waiting list
                    taint = Taint('reg', reg=reg_dep)
                    taints_waitinglist[taint] = Taint('mem', addr=current_run.addr, stmt_id=stmt_idx, mem_addr=action.addr.ast) # TODO: mem_size

                for tmp_dep in action.data.tmp_deps:
                    tmp_taint_set.add(tmp_dep)
                    # Create a new tmp taint, and put it into the waiting list
                    taint = Taint('tmp', tmp=tmp_dep)
                    taints_waitinglist[taint] = Taint('mem', addr=current_run.addr, stmt_id=stmt_idx, mem_addr=action.addr.ast) # TODO: mem_size

                return True

            else:
                return False

    #
    # Helper functions
    #

    @staticmethod
    def _last_branching_statement(statements):
        '''
        Search for the last branching exit, just like
        #   if (t12) { PUT(184) = 0xBADF00D:I64; exit-Boring }
        and then taint the temp variable inside if predicate
        '''
        cmp_stmt_id = None
        cmp_tmp_id = None
        all_statements = len(statements)
        statements = reversed(statements)
        for stmt_rev_idx, stmt in enumerate(statements):
            stmt_idx = all_statements - stmt_rev_idx - 1
            actions = stmt.actions
            # Ugly implementation here
            has_code_action = False
            for a in actions:
                if isinstance(a, simuvex.SimActionExit):
                    has_code_action = True
                    break
            if has_code_action:
                readtmp_action = next(ifilter(lambda r: r.type == 'tmp' and r.action == 'read', actions), None)
                if readtmp_action is not None:
                    cmp_tmp_id = readtmp_action.tmp
                    cmp_stmt_id = stmt_idx
                    break
                else:
                    raise AngrBackwardSlicingError("ReadTempAction is not found. Please report to Fish.")

        return cmp_stmt_id, cmp_tmp_id

import pyvex
import archinfo

from .cfg import CFGNode

#
# Ignored tainted registers. We assume those registers are not tainted, or we just don't care about them
#

# TODO: Add support for more architectures

def _regs_to_offsets(arch_name, regs):
    offsets = set()
    arch = archinfo.arch_from_id(arch_name)
    for r in regs:
        offsets.add(arch.registers[r][0])
    return offsets

_ignored_registers = {
    'X86': { 'esp', 'eip' },
    'AMD64': { 'rsp', 'rip', 'fs' }
}

ignored_registers = { }
for k, v in _ignored_registers.iteritems():
    ignored_registers[k] = _regs_to_offsets(k, v)



from collections import defaultdict
import logging

import networkx
import simuvex
import claripy

from ..entry_wrapper import EntryWrapper, CallStack
from .cfg_base import CFGBase
from ..analysis import Analysis
from ..errors import AngrVFGError, AngrVFGRestartAnalysisNotice, AngrError

l = logging.getLogger(name="angr.analyses.vfg")

# The maximum tracing times of a basic block before we widen the results
MAX_ANALYSIS_TIMES_WITHOUT_MERGING = 10
MAX_ANALYSIS_TIMES = 20

class VFGNode(object):
    def __init__(self, addr, key, state=None):
        self.key = key
        self.addr = addr
        self.state = None
        self.widened_state = None
        self.narrowing_times = 0
        self.all_states  = [ ]
        self.events = [ ]
        self.input_variables = [ ]
        self.actions = [ ]

        if state:
            self.all_states.append(state)
            self.state = state

    def __hash__(self):
        return hash(self.key)

    def __repr__(self):
        s = "VFGNode[0x%x] <%s>" % (self.addr, ", ".join([ (("0x%x" % k) if k else "None") for k in self.key ]))
        return s

    def append_state(self, s, is_widened_state=False):
        """
        Appended a new state to this VFGNode.
        :param s: The new state to append
        :param is_widened_state: Whether it is a widened state or not.
        """

        if not is_widened_state:
            self.all_states.append(s)
            self.state = s

        else:
            self.widened_state = s

class VFG(Analysis):
    '''
    This class represents a control-flow graph with static analysis result.
    '''

    def __init__(self, cfg=None, context_sensitivity_level=2, function_start=None, interfunction_level=0, initial_state=None, avoid_runs=None):
        '''
        :param project: The project object.
        :param context_sensitivity_level: The level of context-sensitivity of this VFG.
                                        It ranges from 0 to infinity. Default 2.
        :param function_start: The address of the function to analyze. N
        :param interfunction_level: The level of interfunction-ness to be
        :param initial_state: A state to use as the initial one
        :param avoid_runs: A list of runs to avoid
        :return:
        '''

        # Related CFG.
        # We can still perform analysis if you don't specify a CFG. But providing a CFG may give you better result.
        self._cfg = cfg

        # Where to start the analysis
        self._start = function_start

        # Other parameters
        self._avoid_runs = [ ] if avoid_runs is None else avoid_runs
        self._context_sensitivity_level = context_sensitivity_level
        self._interfunction_level = interfunction_level

        # Containers
        self._graph = None # TODO: Maybe we want to remove this line?
        self._nodes = None

        self._project = self._p

        self._normal_states = { } # Last available state for each program point without widening
        self._widened_states = { } # States on which widening has occurred

        # Initial states for start analyzing different functions
        # It maps function key to its states
        self._function_initial_states = defaultdict(dict)

        # All final states are put in this list
        self.final_states = [ ]

        self._uninitialized_access = { }
        self._state_initialization_map = defaultdict(list)

        self._state_ignored_variables = { }

        # Begin VFG construction!
        self._construct(initial_state=initial_state)

        self.result = {
            "graph": self._graph,
            "final_states": self.final_states
        }

    def copy(self):
        new_vfg = VFG(self._project)
        new_vfg._cfg = self._cfg
        new_vfg._graph = networkx.DiGraph(self._graph)
        new_vfg._nodes = self._nodes.copy()
        new_vfg._edge_map = self._edge_map.copy()
        return new_vfg

    def _prepare_state(self, function_start, initial_state, function_key):
        # Crawl the binary, create CFG and fill all the refs inside project!
        if initial_state is None:
            if function_start not in self._function_initial_states:
                # We have never saved any initial states for this function
                # Gotta create a fresh state for it
                s = self._project.factory.blank_state(mode="static",
                                              add_options={ simuvex.o.FRESHNESS_ANALYSIS }
                )

                if function_start != self._project.loader.main_bin.entry:
                    # This function might have arguments passed on stack, so make
                    # some room for them.
                    # TODO: Decide the number of arguments and their positions
                    #  during CFG analysis
                    sp = s.registers.load('sp')
                    # Set the address mapping
                    sp_val = s.se.any_int(sp) # FIXME: What will happen if we lose track of multiple sp values?
                    s.memory.set_stack_address_mapping(sp_val,
                                                       s.memory.stack_id(function_start) + '_pre',
                                                       0x0)
                    new_sp = sp - 160
                    s.registers.store('sp', new_sp)
            else:
                if function_key is None:
                    l.debug('We should combine all existing states for this function, then analyze it.')
                    merged_state = None
                    for state in self._function_initial_states[function_start].values():
                        if merged_state is None:
                            merged_state = state
                        else:
                            merged_state, _, _ = merged_state.merge(state)
                    s = merged_state
                elif function_key in self._function_initial_states[function_start]:
                    l.debug('Loading previously saved state for function 0x%x %s', function_start,
                            CallStack.stack_suffix_to_string(function_key))
                    s = self._function_initial_states[function_start][function_key]
                else:
                    raise AngrVFGError('Initial state for function 0x%x and function key %s is not found.' %
                                       (function_start, CallStack.stack_suffix_to_string(function_key)))
        else:
            if function_key is not None:
                # Warn the user
                l.warning('Arguments "function_key" and "initial_state" should not be specified together. ' +
                          'Using specified initial_state as the state.')
            s = initial_state

        # Set the stack address mapping for the initial stack
        s.memory.set_stack_size(s.arch.stack_size)
        initial_sp = s.se.any_int(s.registers.load('sp')) # FIXME: This is bad, as it may lose tracking of multiple sp values
        initial_sp -= s.arch.bits / 8
        s.memory.set_stack_address_mapping(initial_sp,
                                           s.memory.stack_id(function_start),
                                           function_start)

        return s

    def _construct(self, initial_state=None):
        """
        Perform abstract intepretation analysis starting from the given function address. The output is an invariant at
        the beginning (or the end) of each basic block.

        Steps:
        # Generate a CFG first if CFG is not provided.
        # Identify all merge points (denote the set of merge points as Pw) in the CFG.
        # Cut those loop back edges (can be derived from Pw) so that we gain an acyclic CFG.
        # Identify all variables that are 1) from memory loading 2) from initial values, or 3) phi functions. Denote
            the set of those variables as S_{var}.
        # Start real AI analysis and try to compute a fix point of each merge point. Perfrom widening/narrowing only on
            variables \in S_{var}.
        """

        start = self._start if self._start is not None else self._p.entry

        if not self._cfg:
            # Generate a CFG if no CFG is provided
            l.debug("Generating a CFG starting at 0x%x", start)
            self._cfg = self._p.analyses.CFG(context_sensitivity_level=self._context_sensitivity_level,
                starts=(start, )
            )

        cfg = self._cfg

        # Identify all program points to perform a widening
        self._widen_points = cfg.get_loop_back_edges()

        # Cut the loops
        # TODO: I'm directly modifying the CFG. Is it OK, or we should make a copy of it first?
        #for src, dst in merge_points:
        #    self._cfg.graph.remove_edge(src, dst)

        restart_analysis = True

        while restart_analysis:

            restart_analysis = False

            # TODO: Remove those lines
            # Initialization
            # self._normal_states = { } # Last available state for each program point without widening
            # self._widened_states = { } # States on which widening has occurred

            self._function_initial_states = defaultdict(dict)

            # Clear the nodes
            self._nodes = { }

            self._events = [ ]

            try:
                self._ai_analyze(initial_state)
            except AngrVFGRestartAnalysisNotice:
                l.info("Restarting analysis.")
                restart_analysis = True

    def _ai_analyze(self, initial_state=None, function_key=None):
        """
        Construct the value-flow graph, starting at a specific start, until we come to a fixpoint for each merge point.
        """

        # Traverse all the IRSBs, and put them to a dict
        # It's actually a multi-dict, as each SIRSB might have different states
        # on different call predicates
        if self._nodes is None:
            self._nodes = { }

        function_start = self._start
        if function_start is None:
            function_start = self._project.loader.main_bin.entry
        l.debug("Starting from 0x%x", function_start)

        # Prepare the state
        loaded_state = self._prepare_state(function_start, initial_state, function_key)
        loaded_state.ip = function_start
        loaded_state.uninitialized_access_handler = self._uninitialized_access_handler

        loaded_state = loaded_state.arch.prepare_state(loaded_state,
                                                         {'current_function': function_start, }
        )

        # Create the initial path
        # Also we want to identify all fresh variables at each merge point
        entry_state = loaded_state.copy()
        entry_state.options.add(simuvex.o.FRESHNESS_ANALYSIS)
        entry_path = self._project.factory.path(entry_state)
        entry_wrapper = EntryWrapper(entry_path, self._context_sensitivity_level)

        # Initialize a worklist
        worklist = [ entry_wrapper ]

        # Counting how many times a basic block has been traced
        tracing_times = defaultdict(int)

        # For each call, we are always getting two exits: an Ijk_Call that
        # stands for the real call exit, and an Ijk_Ret that is a simulated exit
        # for the retn address. There are certain cases that the control flow
        # never returns to the next instruction of a callsite due to
        # imprecision of the concrete execution. So we save those simulated
        # exits here to increase our code coverage. Of course the real retn from
        # that call always precedes those "fake" retns.
        # Tuple --> (Initial state, call_stack, bbl_stack)
        fake_func_return_paths = { }
        # A dict to log edges and the jumpkind between each basic block
        self._edge_map = defaultdict(list)
        exit_targets = self._edge_map
        # A dict to record all blocks that returns to a specific address
        retn_target_sources = defaultdict(list)

        # Iteratively analyze every exit
        while len(worklist):
            entry_wrapper = worklist.pop()

            # Process the popped path
            self._handle_entry(entry_wrapper, worklist,
                              exit_targets, fake_func_return_paths,
                              tracing_times, retn_target_sources
                              )

            while len(worklist) == 0 and len(fake_func_return_paths) > 0:
                # We don't have any paths remaining. Let's pop a previously-missing return to
                # process
                fake_exit_tuple = fake_func_return_paths.keys()[0]
                fake_exit_state, fake_exit_call_stack, fake_exit_bbl_stack = \
                    fake_func_return_paths.pop(fake_exit_tuple)
                fake_exit_addr = fake_exit_tuple[len(fake_exit_tuple) - 1]
                # Let's check whether this address has been traced before.
                targets = filter(lambda r: r == fake_exit_tuple,
                                 exit_targets)
                if len(targets) > 0:
                    # That block has been traced before. Let's forget about it
                    l.debug("Target 0x%08x has been traced before." +
                            "Trying the next one...",
                            fake_exit_addr)
                    continue

                new_path = self._project.factory.path(fake_exit_state)
                new_path_wrapper = EntryWrapper(new_path,
                                                  self._context_sensitivity_level,
                                                  call_stack=fake_exit_call_stack,
                                                  bbl_stack=fake_exit_bbl_stack)
                worklist.append(new_path_wrapper)
                l.debug("Tracing a missing return 0x%08x, %s", fake_exit_addr,
                        "->".join([hex(i) for i in fake_exit_tuple if i is not None]))
                break

        # Create the real graph
        new_graph = self._create_graph(return_target_sources=retn_target_sources)
        if self._graph is None:
            self._graph = new_graph
        else:
            self._graph.add_edges_from(new_graph.edges(data=True))

        # TODO: Determine the last basic block

        for n in self._graph.nodes():
            if self._graph.out_degree(n) == 0:
                # TODO: Fix the issue when n.successors is empty
                if self._graph.successors(n):
                    self.final_states.extend([ i.state for i in self._graph.successors(n) ])
                else:
                    self.final_states.append(n.state)

    def _create_graph(self, return_target_sources=None):
        '''
        Create a DiGraph out of the existing edge map.
        :param return_target_sources: Used for making up those missing returns
        :return: A networkx.DiGraph() object
        '''
        exit_targets = self._edge_map

        if return_target_sources is None:
            # We set it to a defaultdict in order to be consistent with the
            # actual parameter.
            return_target_sources = defaultdict(list)

        cfg = networkx.DiGraph()
        # The corner case: add a node to the graph if there is only one block
        if len(self._nodes) == 1:
            cfg.add_node(self._nodes[self._nodes.keys()[0]])

        # Adding edges
        for tpl, targets in exit_targets.items():
            basic_block = self._nodes[tpl] # Cannot fail :)
            for ex, jumpkind in targets:
                if ex in self._nodes:
                    target_bbl = self._nodes[ex]
                    cfg.add_edge(basic_block, target_bbl, jumpkind=jumpkind)

                    # Add edges for possibly missing returns
                    if basic_block.addr in return_target_sources:
                        for src_irsb_key in \
                                return_target_sources[basic_block.addr]:
                            cfg.add_edge(self._nodes[src_irsb_key],
                                               basic_block, jumpkind="Ijk_Ret")
                else:
                    # Debugging output
                    def addr_formalize(addr):
                        if addr is None:
                            return "None"
                        else:
                            return "0x%08x" % addr

                    s = "(["
                    for addr in ex[:-1]:
                        s += addr_formalize(addr) + ", "
                    s += "] %s)" % addr_formalize(ex[-1])
                    l.warning("Key %s does not exist.", s)

        return cfg

    def _get_simrun(self, state, current_path, addr):
        error_occured = False
        restart_analysis = False

        jumpkind = 'Ijk_Boring'
        if current_path.state.scratch.jumpkind:
            jumpkind = current_path.state.scratch.jumpkind

        try:
            sim_run = self._project.factory.sim_run(current_path.state, jumpkind=jumpkind)
        except simuvex.SimUninitializedAccessError as ex:
            l.error("Found an uninitialized access (used as %s) at expression %s.", ex.expr_type, ex.expr)
            self._add_expression_to_initialize(ex.expr, ex.expr_type)
            sim_run = None
            error_occured = True
            restart_analysis = True
        except simuvex.SimIRSBError as ex:
            # It's a tragedy that we came across some instructions that VEX
            # does not support. I'll create a terminating stub there
            l.error("SimIRSBError occurred(%s). Creating a PathTerminator.", ex)
            error_occured = True
            sim_run = \
                simuvex.procedures.SimProcedures["stubs"]["PathTerminator"](
                    state, addr=addr)
        except claripy.ClaripyError as ex:
            l.error("ClaripyError: ", exc_info=True)
            error_occured = True
            # Generate a PathTerminator to terminate the current path
            sim_run = \
                simuvex.procedures.SimProcedures["stubs"]["PathTerminator"](
                    state, addr=addr)
        except simuvex.SimError as ex:
            l.error("SimError: ", exc_info=True)

            error_occured = True
            # Generate a PathTerminator to terminate the current path
            sim_run = \
                simuvex.procedures.SimProcedures["stubs"]["PathTerminator"](
                    state, addr=addr)
        except AngrError as ex:
            #segment = self._project.loader.main_bin.in_which_segment(addr)
            l.error("AngrError %s when creating SimRun at 0x%x",
                    ex, addr)
            # We might be on a wrong branch, and is likely to encounter the
            # "No bytes in memory xxx" exception
            # Just ignore it
            error_occured = True
            sim_run = None

        return sim_run, error_occured, restart_analysis

    def _remove_pending_return(self, entry_wrapper, pending_returns):
        """
        Remove all pending returns that are related to the current entry.
        """

        # Build the tuples that we want to remove from the dict fake_func_retn_exits
        tpls_to_remove = [ ]
        call_stack_copy = entry_wrapper.call_stack_copy()
        while call_stack_copy.get_ret_target() is not None:
            ret_target = call_stack_copy.get_ret_target()
            # Remove the current call stack frame
            call_stack_copy.ret(ret_target)
            call_stack_suffix = call_stack_copy.stack_suffix(self._context_sensitivity_level)
            tpl = call_stack_suffix + (ret_target,)
            tpls_to_remove.append(tpl)

        # Remove those tuples from the dict
        for tpl in tpls_to_remove:
            if tpl in pending_returns:
                del pending_returns[tpl]
                l.debug("Removed (%s) from FakeExits dict.",
                        ",".join([hex(i) if i is not None else 'None' for i in tpl]))

    def _handle_entry(self, entry_wrapper, remaining_entries, exit_targets,
                     pending_returns, tracing_times, retn_target_sources):
        '''
        Handles an entry in the program.

        In static mode, we create a unique stack region for each function, and
        normalize its stack pointer to the default stack offset.
        '''

        #
        # Extract initial values
        #
        avoid_runs = self._avoid_runs

        current_path = entry_wrapper.path
        call_stack_suffix = entry_wrapper.call_stack_suffix()
        current_function_address = entry_wrapper.current_function_address
        # We want to narrow the state if widening has occurred before
        is_narrowing = entry_wrapper.is_narrowing

        addr = current_path.addr
        input_state = current_path.state
        simrun_key = call_stack_suffix + (addr,)

        # Initialize the state with necessary values
        self._initialize_state(input_state)

        if simrun_key not in self._nodes:
            vfg_node = VFGNode(addr, simrun_key, state=input_state)
            self._nodes[simrun_key] = vfg_node

        else:
            # Adding a new VFGNode to our nodes dict
            # TODO:

            vfg_node = self._nodes[simrun_key]

        #if widening_stage != 1:
        #    self._normal_states[simrun_key] = input_state

        if vfg_node.widened_state is None and not is_narrowing:
            # This is where merging happens
            merging_occurred, new_input_state = self._handle_states_merging(vfg_node, addr, input_state, tracing_times)

            if new_input_state is None:
                # This basic block doesn't need to be analyzed anymore...
                return

        else:
            if vfg_node.narrowing_times < 5:
                # However, we do want to narrow it if it's a widened state
                # The way we implement narrowing is quite naive: we just reexecute all reachable blocks with this new state
                # for some times, then take the last result
                if vfg_node.widened_state and vfg_node.narrowing_times == 0: new_input_state = vfg_node.widened_state
                else: new_input_state = input_state
                vfg_node.narrowing_times += 1
                is_narrowing = True

            else:
                return

        input_state = new_input_state
        current_path.state = input_state

        # Execute this basic block with input state, and get a new SimRun object
        simrun, error_occured, restart_analysis = self._get_simrun(input_state, current_path, addr)

        if restart_analysis:
            # We should restart the analysis because of something must be changed in the very initial state
            raise AngrVFGRestartAnalysisNotice()

        if simrun is None:
            # Ouch, we cannot get the simrun for some reason
            return

        if addr not in avoid_runs:
            # Obtain successors
            all_successors = simrun.successors + simrun.unconstrained_successors
        else:
            all_successors = [ ]

        # Get ignored variables
        # TODO: We should merge it with existing ignored_variable set!

        if isinstance(simrun, simuvex.SimIRSB) and simrun.default_exit is not None:
            if simrun.default_exit.scratch.ignored_variables:
                self._state_ignored_variables[addr] = simrun.default_exit.scratch.ignored_variables.copy()
            else:
                self._state_ignored_variables[addr] = None
        elif all_successors:
            # This is a SimProcedure instance
            self._state_ignored_variables[addr] = all_successors[0].scratch.ignored_variables.copy() \
                if all_successors[0].scratch.ignored_variables is not None else None

        # Update thumb_addrs. TODO: Do we need it in VFG?
        if isinstance(simrun, simuvex.SimIRSB) and current_path.state.thumb:
            self._thumb_addrs.update(simrun.imark_addrs())

        if len(all_successors) == 0:
            if isinstance(simrun,
                simuvex.procedures.SimProcedures["stubs"]["PathTerminator"]):
                # If there is no valid exit in this branch and it's not
                # intentional (e.g. caused by a SimProcedure that does not
                # do_return) , we should make it return to its callsite.
                # However, we don't want to use its state as it might be
                # corrupted. Just create a link in the exit_targets map.
                retn_target = entry_wrapper.call_stack.get_ret_target()
                if retn_target is not None:
                    new_call_stack = entry_wrapper.call_stack_copy()
                    exit_target_tpl = new_call_stack.stack_suffix(self._context_sensitivity_level) + (retn_target,)
                    exit_targets[call_stack_suffix + (addr,)].append(
                        (exit_target_tpl, 'Ijk_Ret'))
            else:
                # This is intentional. We shall remove all the pending returns generated before along this path.
                self._remove_pending_return(entry_wrapper, pending_returns)

        # If this is a call exit, we shouldn't put the default exit (which
        # is artificial) into the CFG. The exits will be Ijk_Call and
        # Ijk_Ret, and Ijk_Call always goes first
        is_call_jump = any([ self._is_call_jump(i.scratch.jumpkind) for i in all_successors ])
        call_targets = [ i.se.exactly_int(i.ip) for i in all_successors if self._is_call_jump(i.scratch.jumpkind) ]
        call_target = None if not call_targets else call_targets[0]

        is_return_jump = len(all_successors) and all_successors[0].scratch.jumpkind == 'Ijk_Ret'

        # For debugging purpose!
        _dbg_exit_status = { }

        for suc_state in all_successors:

            _dbg_exit_status[suc_state] = ""

            self._handle_successor(suc_state, entry_wrapper, all_successors, is_call_jump, is_return_jump, call_target,
                                   current_function_address, call_stack_suffix, retn_target_sources, pending_returns,
                                   tracing_times, remaining_entries, exit_targets, is_narrowing, _dbg_exit_status)

        # Debugging output
        function_name = self._project.loader.find_symbol_name(simrun.addr)
        module_name = self._project.loader.find_module_name(simrun.addr)

        l.debug("Basic block %s %s", simrun, "->".join([hex(i) for i in call_stack_suffix if i is not None]))
        l.debug("(Function %s of binary %s)", function_name, module_name)
        l.debug("|    Has simulated retn: %s", is_call_jump)
        for suc_state in all_successors:
            if is_call_jump and suc_state.scratch.jumpkind == "Ijk_Ret":
                exit_type_str = "Simulated Ret"
            else:
                exit_type_str = "-"
            try:
                l.debug("|    target: %s %s [%s] %s", hex(suc_state.se.exactly_int(suc_state.ip)), _dbg_exit_status[suc_state], exit_type_str, suc_state.scratch.jumpkind)
            except simuvex.SimValueError:
                l.debug("|    target cannot be concretized. %s [%s] %s", _dbg_exit_status[suc_state], exit_type_str, suc_state.scratch.jumpkind)
        l.debug("len(remaining_exits) = %d, len(fake_func_retn_exits) = %d", len(remaining_entries), len(pending_returns))

    def _handle_states_merging(self, node, addr, new_state, tracing_times):
        """
        Examine if we have reached to a fix point for the current node, and perform merging/widening if necessary.

        :param node: An instance of VFGNode.
        :param new_state: The new input state that we want to compare against.
        :return: A bool value indicating whether we have reached fix point, and the merge state/original state if possible.
        """
        tracing_times[node] += 1

        tracing_count = tracing_times[node]

        l.debug("Analyzing %s for the %dth time", node, tracing_count)

        if tracing_count == 1:
            node.append_state(new_state)
            return False, new_state

        if tracing_count > MAX_ANALYSIS_TIMES:
            l.debug("%s has been analyzed too many times. Skip.", node)

            return False, None

        # Extract two states
        old_state = node.state

        # The widening flag
        widening_occurred = False

        if addr in self._state_ignored_variables:
            old_state.scratch.ignored_variables = self._state_ignored_variables[addr]
        else:
            old_state.scratch.ignored_variables = simuvex.SimVariableSet()

        if addr in set([dst.addr for (src, dst) in self._widen_points]):
            # We reached a merge point

            if tracing_count >= MAX_ANALYSIS_TIMES_WITHOUT_MERGING:

                if node.widened_state is not None:
                    # We want to narrow the state
                    widened_state = node.widened_state
                    merged_state, narrowing_occurred = self._narrow_states(node, old_state, new_state, widened_state)
                    merging_occurred = narrowing_occurred

                else:
                    # We want to widen the state
                    # ... but, we should merge them first
                    merged_state, merging_occurred = self._merge_states(old_state, new_state)
                    # ... then widen it
                    merged_state, widening_occurred = self._widen_states(old_state, merged_state)

                    merging_occurred = widening_occurred

            else:
                # We want to merge them
                merged_state, merging_occurred = self._merge_states(old_state, new_state)

        else:
            # Not a merge point
            # Always merge the state with existing states
            merged_state, merging_occurred = self._merge_states(old_state, new_state)

        if widening_occurred:
            node.append_state(merged_state, is_widened_state=True)

        else:
            node.append_state(merged_state)

        if merging_occurred:
            l.debug("Merging/widening/narrowing occured for %s. Returning a new state.", node)

            return True, merged_state
        else:
            # if simuvex.s_options.WIDEN_ON_MERGE in merged_state.options:
            #    merged_state.options.remove(simuvex.s_options.WIDEN_ON_MERGE)
            l.debug("%s reached fixpoint.", node)

            return False, None

    def _handle_successor(self, suc_state, entry_wrapper, all_successors, is_call_jump, is_return_jump, call_target,
                          current_function_address, call_stack_suffix, retn_target_sources, pending_returns,
                          tracing_count, remaining_entries, exit_targets, is_narrowing, _dbg_exit_status):

        #
        # Extract initial values
        #
        addr = entry_wrapper.path.addr
        jumpkind = suc_state.scratch.jumpkind
        se = suc_state.se

        self._events.extend([ i for i in suc_state.log.events if not isinstance(i, simuvex.SimAction) ])

        #
        # Get instruction pointer
        #
        try:
            if is_return_jump:
                # FIXME: This is a bad practice...
                ret_target = entry_wrapper.call_stack.get_ret_target()
                if ret_target is None:
                    # We have no where to go according to our call stack
                    # However, we still store the state as it is probably the last available state of the analysis
                    call_stack_suffix = entry_wrapper.call_stack_suffix()
                    simrun_key = call_stack_suffix + (addr, )
                    # self._normal_states[simrun_key] = suc_state
                    return

                suc_state.ip = ret_target

            if len(suc_state.se.any_n_int(suc_state.ip, 2)) > 1:
                if is_return_jump:
                    # It might be caused by state merging
                    # We may retrieve the correct ip from call stack
                    suc_state.ip = entry_wrapper.call_stack.get_ret_target()

                else:
                    # Currently we assume a legit jumping target cannot have more than 256 concrete values
                    MAX_NUMBER_OF_CONCRETE_VALUES = 256

                    all_possible_ips = suc_state.se.any_n_int(suc_state.ip, MAX_NUMBER_OF_CONCRETE_VALUES + 1)

                    if len(all_possible_ips) > MAX_NUMBER_OF_CONCRETE_VALUES:
                        l.warning("IP can be concretized to more than %d values, which means it might be corrupted.",
                                  MAX_NUMBER_OF_CONCRETE_VALUES)

                    else:
                        # Call this function for each possible IP
                        for ip in all_possible_ips:
                            suc_state_ = suc_state.copy()
                            suc_state_.ip = ip

                            self._handle_successor(
                                suc_state_, entry_wrapper, all_successors, is_call_jump, is_return_jump, call_target,
                                current_function_address, call_stack_suffix, retn_target_sources, pending_returns,
                                tracing_count, remaining_entries, exit_targets, is_narrowing, _dbg_exit_status)
                    return

            successor_ip = suc_state.se.exactly_int(suc_state.ip)
        except simuvex.SimValueError:
            # TODO: Should fall back to reading targets from CFG
            # It cannot be concretized currently. Maybe we could handle
            # it later, maybe it just cannot be concretized
            return

        # Make a copy of the state in case we use it later
        successor_state = suc_state.copy()

        # Try to remove the unnecessary fakeret successor
        fakeret_successor = None
        if self._is_call_jump(jumpkind):
            fakeret_successor = all_successors[-1]

            # Check if that function is returning
            if self._cfg is not None:
                func = self._cfg.function_manager.function(call_target)
                if func is not None and not func.has_return and len(all_successors) == 2:
                    # Remove the fake return as it is not returning anyway...
                    del all_successors[-1]
                    fakeret_successor = None

        if self._is_call_jump(jumpkind) and \
                len(entry_wrapper.call_stack) > self._interfunction_level:
            l.debug('We are not tracing into a new function 0x%x because we run out of energy :-(', successor_ip)
            # However, we do want to save out the state here
            self._save_state(entry_wrapper, successor_ip, successor_state)

            # Go on to handle the next exit
            return

        # Create a new call stack
        new_call_stack = self._create_callstack(entry_wrapper, successor_ip, jumpkind, is_call_jump, fakeret_successor)
        if new_call_stack is None:
            l.debug("Cannot create a new callstack for address 0x%x", successor_ip)
            return
        new_call_stack_suffix = new_call_stack.stack_suffix(self._context_sensitivity_level)
        new_tpl = new_call_stack_suffix + (successor_ip, )

        # Generate the new BBL stack of target block
        new_bbl_stack = self._create_bblstack(entry_wrapper, jumpkind, successor_ip, is_call_jump, call_stack_suffix,
                                              new_call_stack_suffix, current_function_address)

        # Generate new exits
        if jumpkind == "Ijk_Ret" and not is_call_jump:
            # This is the real retn exit
            # Remember this retn!
            retn_target_sources[successor_ip].append(call_stack_suffix + (addr, ))
            # Check if this retn is inside our fake_func_retn_exits set
            if new_tpl in pending_returns:
                del pending_returns[new_tpl]

        if jumpkind == "Ijk_Ret" and is_call_jump:
            # This is the default "fake" return successor generated at each
            # call. Save them first, but don't process them right
            # away

            # Clear the useless values (like return addresses, parameters) on stack if needed
            if self._cfg is not None:
                current_function = self._cfg.function_manager.function(call_target)
                if current_function is not None:
                    sp_difference = current_function.sp_delta
                else:
                    sp_difference = 0
                reg_sp_offset = successor_state.arch.sp_offset
                reg_sp_expr = successor_state.registers.load(reg_sp_offset) + sp_difference
                successor_state.registers.store(successor_state.arch.sp_offset, reg_sp_expr)

                # Clear the return value with a TOP
                top_si = successor_state.se.TopStridedInterval(successor_state.arch.bits)
                successor_state.registers.store(successor_state.arch.ret_offset, top_si)

                pending_returns[new_tpl] = \
                    (successor_state, new_call_stack, new_bbl_stack)
                _dbg_exit_status[suc_state] = "Appended to fake_func_retn_exits"

        else:
            successor_path = self._project.factory.path(successor_state)
            if simuvex.o.ABSTRACT_MEMORY in suc_state.options:
                if self._is_call_jump(suc_state.scratch.jumpkind):
                    # If this is a call, we create a new stack address mapping
                    reg_sp_si = self._create_stack_region(successor_path.state, successor_path.addr)

                    # Save the new sp register
                    new_reg_sp_expr = successor_path.state.se.ValueSet(bits=suc_state.arch.bits)
                    new_reg_sp_expr.model.set_si('global', reg_sp_si.copy())
                    reg_sp_offset = successor_state.arch.sp_offset
                    successor_path.state.registers.store(reg_sp_offset, new_reg_sp_expr)

                elif suc_state.scratch.jumpkind == "Ijk_Ret":
                    # Remove the existing stack address mapping
                    # FIXME: Now we are assuming the sp is restored to its original value
                    reg_sp_offset = successor_path.state.arch.sp_offset
                    reg_sp_expr = successor_path.state.registers.load(reg_sp_offset).model

                    reg_sp_si = reg_sp_expr.items()[0][1]
                    reg_sp_val = reg_sp_si.min
                    # TODO: Finish it!

            new_exit_wrapper = EntryWrapper(successor_path,
                                            self._context_sensitivity_level,
                                            call_stack=new_call_stack,
                                            bbl_stack=new_bbl_stack,
                                            is_narrowing=is_narrowing)
            r = self._append_to_remaining_entries(remaining_entries, new_exit_wrapper)
            _dbg_exit_status[suc_state] = r

        if not is_call_jump or jumpkind != "Ijk_Ret":
            exit_targets[call_stack_suffix + (addr,)].append((new_tpl, jumpkind))
        else:
            # This is the fake return!
            exit_targets[call_stack_suffix + (addr,)].append((new_tpl, "Ijk_FakeRet"))

    def _is_call_jump(self, jumpkind):
        if jumpkind == 'Ijk_Call' or jumpkind.startswith('Ijk_Sys_'):
            return True
        return False

    def _append_to_remaining_entries(self, remaining_entries, exit_wrapper):
        simrun_key = exit_wrapper.call_stack_suffix() + (exit_wrapper.path.addr, )
        simrun_key = simrun_key[1 : ]

        node = self._cfg.get_node(simrun_key)

        if node is None:
            result = "Appended [not in CFG]"
            remaining_entries.append(exit_wrapper)

        else:
            in_degree = self._cfg.graph.in_degree(node)

            if in_degree <= 1:
                result = "Appended [not a merge point]"
                remaining_entries.append(exit_wrapper)

            else:
                result = "Inserted at front [a merge point]"
                remaining_entries.insert(0, exit_wrapper)

        return result

    def _widen_states(self, old_state, new_state):
        """
        Perform widen operation on the given states, and return a new one.

        :param old_state:
        :param new_state:
        :return: The widened state, and whether widening has occurred
        """

        # print old_state.dbg_print_stack()
        # print new_state.dbg_print_stack()

        if old_state.scratch.ignored_variables is None:
            old_state.scratch.ignored_variables = new_state.scratch.ignored_variables

        widened_state, widening_occurred = old_state.widen(new_state)

        # print "Widened: "
        # print widened_state.dbg_print_stack()

        return widened_state, widening_occurred

    def _narrow_states(self, node, old_state, new_state, previously_widened_state):
        """
        Try to narrow the state!

        :param old_state:
        :param new_state:
        :param previously_widened_state:
        :return: The narrowed state, and whether a narrowing has occurred
        """

        s = previously_widened_state.copy()

        narrowing_occurred = False

        # We will narrow all variables besides those in ignored_variables set

        # Check each fresh variable in new_state, and update them in previously_widened_state if needed.
        # Registers
        ignored_register_vars = self._state_ignored_variables[node.addr].register_variables

        # TODO: Finish the narrowing logic

        return s, narrowing_occurred

    def _merge_states(self, old_state, new_state):
        """
        Merge two given states, and return a new one.

        :param old_state:
        :param new_state:
        :return: The merged state, and whether a merging has occurred
        """

        # print old_state.dbg_print_stack()
        # print new_state.dbg_print_stack()

        if old_state.scratch.ignored_variables is None:
            old_state.scratch.ignored_variables = new_state.scratch.ignored_variables

        merged_state, _, merging_occurred = old_state.merge(new_state)

        # print "Merged: "
        # print merged_state.dbg_print_stack()

        return merged_state, merging_occurred

    def _create_stack_region(self, successor_state, successor_ip):
        reg_sp_offset = successor_state.arch.sp_offset
        reg_sp_expr = successor_state.registers.load(reg_sp_offset)

        if type(reg_sp_expr.model) is claripy.BVV:
            reg_sp_val = successor_state.se.any_int(reg_sp_expr)
            reg_sp_si = successor_state.se.SI(to_conv=reg_sp_expr)
            reg_sp_si = reg_sp_si.model
        elif type(reg_sp_expr.model) in (int, long):
            reg_sp_val = reg_sp_expr.model
            reg_sp_si = successor_state.se.SI(bits=successor_state.arch.bits, to_conv=reg_sp_val)
            reg_sp_si = reg_sp_si.model
        else:
            reg_sp_si = reg_sp_expr.model.items()[0][1]
            reg_sp_val = reg_sp_si.min

        reg_sp_val = reg_sp_val - successor_state.arch.bits / 8  # TODO: Is it OK?
        new_stack_region_id = successor_state.memory.stack_id(successor_ip)
        successor_state.memory.set_stack_address_mapping(reg_sp_val,
                                                        new_stack_region_id,
                                                        successor_ip)

        return reg_sp_si

    def _create_callstack(self, entry_wrapper, successor_ip, jumpkind, is_call_jump, fakeret_successor):
        addr = entry_wrapper.path.addr

        if self._is_call_jump(jumpkind):
            if len(entry_wrapper.call_stack) <= self._interfunction_level:
                new_call_stack = entry_wrapper.call_stack_copy()
                # Notice that in ARM, there are some freaking instructions
                # like
                # BLEQ <address>
                # It should give us three exits: Ijk_Call, Ijk_Boring, and
                # Ijk_Ret. The last exit is simulated.
                # Notice: We assume the last exit is the simulated one
                if fakeret_successor is None:
                    retn_target_addr = None
                else:
                    retn_target_addr = fakeret_successor.se.exactly_n_int(fakeret_successor.ip, 1)[0]

                # Create call stack
                new_call_stack.call(addr, successor_ip,
                                    retn_target=retn_target_addr)
            else:
                return None

        elif jumpkind == "Ijk_Ret" and not is_call_jump:
            new_call_stack = entry_wrapper.call_stack_copy()
            new_call_stack.ret(successor_ip)

        else:
            # Normal control flow transition
            new_call_stack = entry_wrapper.call_stack

        return new_call_stack

    def _create_bblstack(self, entry_wrapper, jumpkind, successor_ip, is_call_jump, call_stack_suffix,
                         new_call_stack_suffix, current_function_address):
        if self._is_call_jump(jumpkind):
            new_bbl_stack = entry_wrapper.bbl_stack_copy()
            new_bbl_stack.call(new_call_stack_suffix, current_function_address)
            new_bbl_stack.push(new_call_stack_suffix, current_function_address, successor_ip)
        elif jumpkind == "Ijk_Ret" and not is_call_jump:
            new_bbl_stack = entry_wrapper.bbl_stack_copy()
            new_bbl_stack.ret(call_stack_suffix, current_function_address)
        else:
            new_bbl_stack = entry_wrapper.bbl_stack_copy()
            new_bbl_stack.push(new_call_stack_suffix, current_function_address, successor_ip)

        return new_bbl_stack

    def _save_state(self, entry_wrapper, successor_ip, successor_state):
        addr = entry_wrapper.path.addr

        new_call_stack_suffix = entry_wrapper.call_stack.stack_suffix(self._context_sensitivity_level)
        function_key = new_call_stack_suffix + (addr, )
        function_addr = successor_ip
        l.debug('Saving out the state for function 0x%x with function_key %s',
                function_addr,
                CallStack.stack_suffix_to_string(function_key))
        if function_addr in self._function_initial_states and \
                        function_key in self._function_initial_states[function_addr]:
            existing_state = self._function_initial_states[function_addr][function_key]
            merged_state, _, _ = existing_state.merge(successor_state)
            self._function_initial_states[function_addr][function_key] = merged_state
        else:
            self._function_initial_states[function_addr][function_key] = successor_state

    def _uninitialized_access_handler(self, mem_id, addrs, length, expr, bbl_addr, stmt_idx):
        if type(addrs) is not list:
            addrs = [ addrs ]

        for addr in addrs:
            if expr.model.name not in self._uninitialized_access:
                self._uninitialized_access[expr.model.name] = (bbl_addr, stmt_idx, mem_id, addr, length)

    def _find_innermost_uninitialized_expr(self, expr):
        result = [ ]

        if not expr.args:
            if hasattr(expr.model, 'uninitialized') and expr.model.uninitialized:
                result.append(expr)

        else:
            tmp_result = [ ]

            for a in expr.args:
                if isinstance(a, claripy.Base):
                    r = self._find_innermost_uninitialized_expr(a)

                    if r:
                        tmp_result.extend(r)

            if tmp_result:
                result = tmp_result

            elif not tmp_result and \
                    hasattr(expr.model, 'uninitialized') and \
                    expr.model.uninitialized:
                result.append(expr)

        return result

    def _add_expression_to_initialize(self, expr, expr_type):

        expr = self._find_innermost_uninitialized_expr(expr)

        next_expr_type = expr_type

        for ex in expr:
            name = ex.model.name

            if name in self._uninitialized_access:
                bbl_addr, stmt_idx, mem_id, addr, length = self._uninitialized_access[name]
                # FIXME: Here we are reusing expr_type for all uninitialized variables.
                # FIXME: However, it will cause problems for the following case:
                # FIXME: addr(uninit) = addr_a(uninit) + offset_b(uninit)
                # FIXME: In this case, both addr_a and offset_b will be treated as addresses
                # FIXME: We should consider a fix for that
                self._state_initialization_map[bbl_addr].append((mem_id, addr, length, next_expr_type, stmt_idx))
                if next_expr_type == 'addr':
                    next_expr_type = 'value'
            else:
                raise Exception('TODO: Please report it to Fish')

    def _initialize_state(self, state):

        se = state.se
        bbl_addr = se.exactly_int(state.ip)

        if bbl_addr in self._state_initialization_map:
            for data in self._state_initialization_map[bbl_addr]:
                mem_id, addr, length, expr_type, stmt_idx = data

                # Initialize it
                if expr_type == 'addr':
                    # Give it an address
                    value = se.VS(region='_init_%x_%d' % (bbl_addr, stmt_idx), bits=state.arch.bits, val=0)

                elif expr_type == 'value':
                    # Give it a value
                    value = se.SI(bits=length, stride=1, lower_bound=-(2**length-1), upper_bound=2**length)

                else:
                    raise Exception('Not implemented. Please report it to Fish.')

                # Write to the state
                if mem_id == 'reg':
                    state.registers.store(addr, value)
                else:
                    # TODO: This is completely untested!
                    region_id, offset = addr
                    target_addr = se.VS(region=region_id, bits=state.arch.bits, val=offset)

                    state.memory.store(target_addr, value, size=length)

    def _get_block_addr(self, b): #pylint:disable=R0201
        if isinstance(b, simuvex.SimIRSB):
            return b.first_imark.addr
        elif isinstance(b, simuvex.SimProcedure):
            return b.addr
        else:
            raise Exception("Unsupported block type %s" % type(b))



import logging
import os.path
import pickle
from collections import namedtuple

from ..analysis import Analysis

STACKFRAME_PREFIX = "stack_"
ARRAY_SIZE_LBOUND = 8  # Number of bytes that are still not regarded as an array

l = logging.getLogger(name="angr.buffer")

class Overlap(namedtuple("Overlap",
                         'offset, size, boundary, state, aloc, block, statement, stackframe, out_of_frame, instruction')):
    # Immutable namedtuple containing info about the Overlap.
    def __new__(cls, aloc, seg, stackframe, state, boundary):
        return super(Overlap, cls).__new__(cls, seg.offset, seg.size, boundary, state, aloc, aloc.basicblock_key,
                                           aloc.statement_id, stackframe, seg.offset + seg.size > 0,
                                           "0x%x:%d" % (aloc.basicblock_key, aloc.statement_id))

    def __unicode__(self):
        return u"Overlap[Stackframe %s, Offset: -0x%x, Size: %d, OutOfFrame: %r, Block: 0x%x, Statement: %d]" % (
            self.stackframe.id, - self.offset, self.size, self.out_of_frame, self.block, self.statement)

    def __repr__(self):
        return unicode(self).encode('utf-8')


def process_vfg(vfg):
    found = []

    for state in vfg.final_states:
        regions = state.memory.regions
        stackframes = [v for k, v in regions.iteritems() if k.startswith(STACKFRAME_PREFIX)]
        if len(stackframes) == 0:
            l.info("No stackframe found in state %s. Skipping.", str(state))
        for stackframe in stackframes:  # Should usually be 1 anyway.
            l.debug("Processing stackframe %s in %s", str(stackframe), str(state))
            all_segments = ((a, s) for a in stackframe.alocs.values() for s in a._segment_list)
            segs = sorted(all_segments, key=lambda (_, seg): seg.offset)

            for idx, (aloc, seg) in enumerate(segs):
                if seg.size > ARRAY_SIZE_LBOUND:  # variable <= 64 bit: unlikely to be an array.
                    boundary = 0  # Offsets are negative
                    for (_, seg2) in segs[idx:]:
                        if seg2.offset != seg.offset:
                            boundary = seg2.offset
                            break

                    if seg.offset + seg.size > boundary:  # TODO: Is this off by 1?
                        overlap = Overlap(aloc, seg, stackframe, state, boundary)
                        l.info("Found overlap: " + str(overlap))
                        found.append(overlap)

    return found


class BufferOverlap(Analysis):
    """
    This class looks for overlapping buffers on the stack and buffer that exceed the stackframe
    """
    def process_function(self, f, interfunction_level):
        interfunction_level = self._interfunction_level if interfunction_level is None else interfunction_level

        with self._resilience():
            self.vfg._construct(f, interfunction_level=interfunction_level)
            self.seeker._construct(func_start=f)

        self.finished_functions.add(f)

    def __init__(self, cfg=None, functions=None):
        """
        This class looks for overlapping buffers on the stack and buffer that exceed the stackframe.
        Exceeding a stackframe is very likely a leak, overlapping buffers might in rare cases be a reused stack region
        :return: a dict containing a list of occuring overflows for every instruction. Example:
        {'0x4004f0:19': [Overlap[..., Offset: -0x58, ...], Overlap[..., Offset:-0x50, ...]]}
        found the same overlap in two possible states, caused by instruction 19 in basic block 0x4004f0.
        """

        self.result = {}
        self._cfg = cfg

        all_functions = functions

        if not all_functions:
            self._cfg = cfg if cfg else self._p.results.CFG
            all_functions = self._cfg.function_manager.functions


        for func in all_functions:
            if self._p.is_hooked(func):
                continue
            # Create one VFG for every function in the binary
            vfg = self._p.analyses.VFG(cfg=self._cfg, function_start=func, interfunction_level=3, context_sensitivity_level=2)
            for overlap in process_vfg(vfg):
                if overlap.instruction not in self.result:
                    self.result[overlap.instruction] = []
                self.result[overlap.instruction].append(overlap)



import logging

from ..surveyor import Surveyor

l = logging.getLogger("angr.surveyors.executor")

class Executor(Surveyor):
    '''
    This class handles pure concrete execution related issues.
    No state splitting is ever allowed.
    '''
    def __init__(self, project, start, final_addr=None, \
                 pickle_paths=None, max_run=50000):
        Surveyor.__init__(self, project, start=start, pickle_paths=pickle_paths)
        self._project = project
        self._final_addr = final_addr
        self._max_run = max_run
        self._done = False
        self._error_occured = False
        self._run_counter = 0

    @property
    def done(self):
        if self.error_occured:
            return True
        if len(self.active) > 1:
            raise Exception("We have more than one path in concrete mode." + \
                            " Something is wrong.")
        elif len(self.active) == 0:
            return True
        else:
            path = self.active[0]
            if path.state is not None and \
                    path.state.se.is_true(path.state.ip == self._final_addr):
                self.deadended.append(self.active[0])
                self.active = []
                return True
        return False

    @property
    def error_occured(self):
        if len(self.active) == 0:
            # The path ends before get to the target
            return True
        elif self._run_counter > self._max_run:
            return True
        return False

    @property
    def last_state(self):
        if self.done or self.error_occured:
            return None
        return self.active[0].state

    def tick(self):
        self._run_counter += 1
        Surveyor.tick(self)

        if len(self.active) > 0:
             l.debug("Ran %d run, %s is active...", self._run_counter, self.active[0].previous_run)
        else:
             l.debug("Ran %d run, no more actives...", self._run_counter)

from . import all_surveyors
all_surveyors['Executor'] = Executor



from collections import defaultdict
import logging

import networkx

l = logging.getLogger("angr.analyses.cdg")

from ..analysis import Analysis

# Control dependency graph

class TempNode(object):
    def __init__(self, label):
        self._label = label

    def __repr__(self):
        return self._label

class CDG(Analysis):
    def __init__(self, cfg=None, start=None):
        self._project = self._p
        self._binary = self._project.loader.main_bin
        self._start = start

        self._cfg = cfg if cfg is not None else self._p.analyses.CFG()
        self._acyclic_cfg = self._cfg.copy()
        # The CFG we use should be acyclic!
        self._acyclic_cfg.remove_cycles()

        self._ancestor = None
        self._semi = None
        self._post_dom = None

        self._cdg = None
        self._label = None
        # Debugging purpose
        if hasattr(self._cfg, "get_node"):
            # FIXME: We should not use get_any_irsb in such a real setting...
            self._entry = self._cfg.get_any_node(self._p.entry)

        self.construct()

    def construct(self):
        # Construct post-dominator tree
        self.pd_construct()
        # l.debug("Post dominators: \n%s", self._post_dom)

        self._cdg = networkx.DiGraph()
        # For each node (A,B), traverse back from B until the parent node of A,
        # and label them as control dependent on A
        for a in self._cfg.nodes():
            # FIXME: Dirty fix!
            if a not in self._post_dom:
                continue

            successors = self._cfg.get_successors(a)
            for b in successors:
                # # FIXME: Dirty fix!
                # if b not in self._post_dom:
                #     continue
                # Let's first check whether A's parent lies on B's path to the root
                dependent_flag = False
                tmp = b
                while tmp != None:
                    if tmp == self._post_dom[a]:
                        dependent_flag = True
                        break
                    tmp = self._post_dom[tmp]

                if self._post_dom[a] != b and dependent_flag:
                    # B doesn't post-dominate A
                    tmp = b
                    while tmp != self._post_dom[a]: # and tmp != None: # FIXME: tmp != None is a dirty fix
                        self._cdg.add_edge(a, tmp) # tmp is dependent on A
                        if b in self._post_dom:
                            tmp = self._post_dom[tmp]
                        else:
                            break

        self._post_process()

    def get_post_dominators(self):
        return self._post_dom

    def _post_process(self):
        '''
        There are cases where a loop has two overlapping loop headers thanks
        to the way VEX is dealing with continuous instructions. As we were
        breaking the connection between the second loop header and its
        successor, we shall restore them in our CDG.
        '''
        # TODO: Verify its correctness
        loop_back_edges = self._cfg.get_loop_back_edges()
        for b1, b2 in loop_back_edges:
            self._cdg.add_edge(b1, b2)

    def get_predecessors(self, run):
        if run in self._cdg.nodes():
            return self._cdg.predecessors(run)
        else:
            return []

    def pd_construct(self):
        normalized_graph, vertices, parent = self.pd_normalize_graph()

        bucket = defaultdict(set)
        dom = [None] * (len(vertices))
        self._ancestor = [None] * (len(vertices) + 1)

        range_ = range(1, len(vertices))
        range_.reverse()
        for i in range_:
            w = vertices[i]
            if w not in parent:
                # It's one of the start nodes
                continue
            predecessors = normalized_graph.predecessors(w)
            for v in predecessors:
                u = self.pd_eval(v)
                if self._semi[u.index].index < self._semi[w.index].index:
                    self._semi[w.index] = self._semi[u.index]
            bucket[vertices[self._semi[w.index].index].index].add(w)
            self.pd_link(parent[w], w)
            for v in bucket[parent[w].index]:
                u = self.pd_eval(v)
                if self._semi[u.index].index < self._semi[v.index].index:
                    dom[v.index] = u
                else:
                    dom[v.index] = parent[w]
            bucket[parent[w].index].clear()

        for i in range(1, len(vertices)):
            w = vertices[i]
            if w not in parent:
                continue
            if dom[w.index] != vertices[self._semi[w.index].index]:
                dom[w.index] = dom[dom[w.index].index]

        self._post_dom = {}
        for i in range(1, len(vertices)):
            self._post_dom[vertices[i]] = dom[i]

        self._pd_post_process()

    def _pd_post_process(self):
        '''
        Take care of those loop headers/tails where we manually broke their
        connection to the next BBL
        '''
        loop_back_edges = self._cfg.get_loop_back_edges()
        for b1, b2 in loop_back_edges:
            # The edge between b1 and b2 is manually broken
            # The post dominator of b1 should be b2 (or not?)
            successors = self._acyclic_cfg.get_successors(b1)
            if len(successors) == 0:
                if b2 in self._post_dom:
                    self._post_dom[b1] = b2
                else:
                    l.debug("%s is not in post dominator dict.", b2)

    def pd_normalize_graph(self):
        # We want to reverse the CFG, and label each node according to its
        # order in a DFS
        graph = networkx.DiGraph()

        n = self._start if self._start is not None else self._entry
        assert n is not None
        queue = [n]
        start_node = TempNode("start_node")
        traversed_nodes = set()
        while len(queue) > 0:
            node = queue.pop()
            traversed_nodes.add(node)
            successors = self._acyclic_cfg.get_successors(node)
            if len(successors) == 0:
                # Add an edge between this node and our start node
                graph.add_edge(start_node, node)
            for s in successors:
                graph.add_edge(s, node) # Reversed
                if s not in traversed_nodes:
                    queue.append(s)

        # Add a start node and an end node
        graph.add_edge(n, TempNode("end_node"))

        all_nodes_count = len(traversed_nodes) + 2 # A start node and an end node
        l.debug("There should be %d nodes in all", all_nodes_count)
        counter = 0
        vertices = ["placeholder"]
        scanned_nodes = set()
        parent = {}
        while True:
            # DFS from the current start node
            stack = [start_node]
            while len(stack) > 0:
                node = stack.pop()
                counter += 1
                node.index = counter
                scanned_nodes.add(node)
                vertices.append(node)
                successors = graph.successors(node)
                for s in successors:
                    if s not in scanned_nodes:
                        stack.append(s)
                        parent[s] = node
                        scanned_nodes.add(s)

            if counter >= all_nodes_count:
                break

            l.debug("%d nodes are left out during the DFS. They must formed a cycle themselves.", all_nodes_count - counter)
            # Find those nodes
            leftovers = [s for s in traversed_nodes if s not in scanned_nodes]
            graph.add_edge(start_node, leftovers[0])
            # We have to start over...
            counter = 0
            parent = {}
            scanned_nodes = set()
            vertices = ["placeholder"]

        self._semi = vertices[::]
        self._label = vertices[::]

        return (graph, vertices, parent)

    def pd_link(self, v, w):
        self._ancestor[w.index] = v

    def pd_eval(self, v):
        if self._ancestor[v.index] is None:
            return v
        else:
            self.pd_compress(v)
            return self._label[v.index]

    def pd_compress(self, v):
        if self._ancestor[self._ancestor[v.index].index] != None:
            self.pd_compress(self._ancestor[v.index])
            if self._semi[self._label[self._ancestor[v.index].index].index].index < self._semi[self._label[v.index].index].index:
                self._label[v.index] = self._label[self._ancestor[v.index].index]
            self._ancestor[v.index] = self._ancestor[self._ancestor[v.index].index]



#!/usr/bin/env python

import logging
l = logging.getLogger("angr.surveyors.slicecutor")

from ..surveyor import Surveyor
from ..errors import AngrExitError

from collections import defaultdict

# pylint: disable=W0212,

#
# HappyGraph is just here for testing. Please ignore it!
#
class HappyGraph(object):
    def __init__(self, path = None, paths=None, strict=False):
        if not strict:
            self.jumps = defaultdict(lambda: False)
        else:
            self.jumps = { }

        if paths is None:
            paths = [ ]
        if path is not None:
            paths.append(path)

        for p in paths:
            for i in range(len(p.addr_backtrace) - 1):
                self.jumps[(p.addr_backtrace[i], p.addr_backtrace[i+1])] = True
            self.jumps[(p.addr_backtrace[-1], p.addr)] = True

        self._merge_points = [ ]

    def filter_path(self, path): # pylint: disable=W0613,R0201,
        return True

    def should_take_exit(self, src, dst): # pylint: disable=W0613,R0201,
        return self.jumps[(src, dst)]

    def get_whitelisted_statements(self, addr): # pylint: disable=W0613,R0201,
        return None

    def get_last_statement_index(self, addr): # pylint: disable=W0613,R0201,
        return None

    def merge_points(self, path): # pylint: disable=W0613,R0201,
        return self._merge_points

    def path_priority(self, path): # pylint: disable=W0613,R0201,
        return 1

class Slicecutor(Surveyor):
    '''The Slicecutor is a surveyor that executes provided code slices.'''

    def __init__(self, project, annotated_cfg, start=None, targets=None, max_concurrency=None, max_active=None, max_loop_iterations=None, pickle_paths=None, merge_countdown=10):
        Surveyor.__init__(self, project, start=start, max_concurrency=max_concurrency, max_active=max_active, pickle_paths=pickle_paths)

        # the loop limiter
        self._max_loop_iterations = max_loop_iterations if max_loop_iterations else None

        # the project we're slicing up!
        self._project = project

        # the annotated cfg to determine what to execute
        self._annotated_cfg = annotated_cfg

        # these are paths that are taking exits that the annotated CFG does not
        # know about
        self.mysteries = [ ]

        # these are paths that we cut due to the slicing
        self.cut = [ ]

        # those that have reached one of our targets
        self.reached_targets = []

        if targets is not None:
            self._targets = targets
        else:
            self._targets = []

        # mergesanity!
        self._merge_candidates = defaultdict(list)
        self._merge_countdowns = { }
        self.merge_countdown = merge_countdown

    def filter_path(self, path):
        l.debug("Checking path %s for filtering...", path)
        if not self._annotated_cfg.filter_path(path):
            l.debug("... %s is cut by AnnoCFG explicitly.", path)
            self.cut.append(self.suspend_path(path))
            return False

        l.debug("... checking loop iteration limit")
        if self._max_loop_iterations is not None and path.detect_loops() > self._max_loop_iterations:
            l.debug("... limit reached")
            return False

        l.debug("... checking if %s should wait for a merge.", path)
        if path.addr in path._upcoming_merge_points:
            l.debug("... it should!")
            if path.addr not in self._merge_candidates:
                self._merge_candidates[path.addr] = [ ]

            self._merge_candidates[path.addr].append(path)
            self._merge_countdowns[path.addr] = self.merge_countdown
            return False

        return True

    def tick_path(self, path):
        path._upcoming_merge_points = self._annotated_cfg.merge_points(path)

        path_successors = Surveyor.tick_path(self, path)
        new_paths = [ ]

        mystery = False
        cut = False

        # No new paths if the current path is already the target
        if not path.errored and path.addr in self._targets:
            self.reached_targets.append(self.suspend_path(path))
            return []

        l.debug("%s ticking path %s, last run is %s", self, path, path.previous_run)
        for successor in path_successors:
            dst_addr = successor.addr
            l.debug("... checking exit to 0x%x from %s", dst_addr, path.previous_run)
            try:
                taken = self._annotated_cfg.should_take_exit(path.addr, dst_addr)
            except AngrExitError: # TODO: which exception?
                l.debug("... annotated CFG did not know about it!")
                mystery = True
                continue

            if taken:
                l.debug("... taking the exit.")
                new_paths.append(successor)
                # the else case isn't here, because the path should set errored in this
                # case and we'll catch it below
            else:
                l.debug("... not taking the exit.")
                cut = True

        if mystery: self.mysteries.append(self.suspend_path(path))
        if cut: self.cut.append(self.suspend_path(path))

        return new_paths

    def pre_tick(self):

        # Set whitelists and last statements
        for p in self.active:
            addr = p.state.se.exactly_n_int(p.state.ip, 1)[0]
            whitelist = self._annotated_cfg.get_whitelisted_statements(addr)
            last_stmt = self._annotated_cfg.get_last_statement_index(addr)
            p.stmt_whitelist = whitelist
            p.last_stmt = last_stmt

        done_addrs = [ ]
        for addr, count in self._merge_countdowns.iteritems():
            l.debug("Checking merge point 0x%x with countdown %d.", addr, count)
            if count == 0:
                to_merge = self._merge_candidates[addr]
                l.debug("... merging %d paths!", len(to_merge))

                if len(to_merge) > 1:
                    new_path = to_merge[0].merge(*(to_merge[1:]))
                else:
                    new_path = to_merge[0]

                new_path.extra_length += self.merge_countdown
                done_addrs.append(addr)
                self.active.append(new_path)
            else:
                self._merge_countdowns[addr] -= 1

        for d in done_addrs:
            del self._merge_candidates[d]
            del self._merge_countdowns[d]

    @property
    def done(self):
        return (len(self.active) + len(self._merge_countdowns)) == 0

    def path_comparator(self, a, b):
        if a.weighted_length != b.weighted_length:
            return b.weighted_length - a.weighted_length
        return a.addr_backtrace.count(a.addr_backtrace[-1]) - b.addr_backtrace.count(b.addr_backtrace[-1])
        #return self._annotated_cfg.path_priority(a) - self._annotated_cfg.path_priority(b)

    def __repr__(self):
        return "<Slicecutor with paths: %s, %d cut, %d mysteries, %d reached targets, %d waiting to merge>" % (Surveyor.__repr__(self), len(self.cut), len(self.mysteries), len(self.reached_targets), sum(len(i) for i in self._merge_candidates.values()))

from . import all_surveyors
all_surveyors['Slicecutor'] = Slicecutor
